(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/admin/system-status-server.js                                //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
var Crypto = Npm.require("crypto");                                    // 1
                                                                       //
var hashSessionId = function (sessionId) {                             // 3
  return Crypto.createHash("sha256").update(sessionId).digest("base64");
};                                                                     //
                                                                       //
Meteor.publish("systemStatus", function () {                           // 7
  // A pseudocollection containing the number of active grain sessions and unique
  // userIds associated with these sessions                            //
  var db = this.connection.sandstormDb;                                // 10
  if (!db.isAdminById(this.userId)) {                                  // 11
    throw new Meteor.Error(403, "User must be admin to view system status.");
  }                                                                    //
                                                                       //
  var _this = this;                                                    // 15
  var userIdToSessionHashes = {}; // Maps userId => list of (sha256(sessionId))
  var grainIdToSessionHashes = {}; // Maps grainId => list of (sha256(sessionId))
                                                                       //
  var userIdCount = 0;                                                 // 19
  var grainIdCount = 0;                                                // 20
                                                                       //
  this.added("systemStatus", "globalStatus", {                         // 22
    activeUsers: userIdCount,                                          // 23
    activeGrains: grainIdCount                                         // 24
  });                                                                  //
                                                                       //
  var query = db.collections.sessions.find();                          // 27
  var handle = query.observe({                                         // 28
    added: function (session) {                                        // 29
      var hashedSessionId = hashSessionId(session._id);                // 30
      var changed = {};                                                // 31
                                                                       //
      // Update userId cache.                                          //
      if (session.userId) {                                            // 34
        if (userIdToSessionHashes[session.userId] === undefined) {     // 35
          userIdToSessionHashes[session.userId] = [];                  // 36
          userIdCount = userIdCount + 1;                               // 37
          changed.activeUsers = userIdCount;                           // 38
        }                                                              //
                                                                       //
        userIdToSessionHashes[session.userId] = _.union(userIdToSessionHashes[session.userId], [hashedSessionId]);
      }                                                                //
                                                                       //
      // Update grainId cache.                                         //
      if (grainIdToSessionHashes[session.grainId] === undefined) {     // 46
        grainIdToSessionHashes[session.grainId] = [];                  // 47
        grainIdCount = grainIdCount + 1;                               // 48
        changed.activeGrains = grainIdCount;                           // 49
      }                                                                //
                                                                       //
      grainIdToSessionHashes[session.grainId] = _.union(grainIdToSessionHashes[session.grainId], [hashedSessionId]);
                                                                       //
      if (!_.isEmpty(changed)) {                                       // 55
        _this.changed("systemStatus", "globalStatus", changed);        // 56
      }                                                                //
    },                                                                 //
                                                                       //
    removed: function (session) {                                      // 60
      var hashedSessionId = hashSessionId(session._id);                // 61
      var changed = {};                                                // 62
                                                                       //
      if (session.userId) {                                            // 64
        userIdToSessionHashes[session.userId] = _.without(userIdToSessionHashes[session.userId], hashedSessionId);
                                                                       //
        if (userIdToSessionHashes[session.userId].length === 0) {      // 68
          userIdToSessionHashes[session.userId] = undefined;           // 69
          userIdCount = userIdCount - 1;                               // 70
          changed.activeUsers = userIdCount;                           // 71
        }                                                              //
      }                                                                //
                                                                       //
      grainIdToSessionHashes[session.grainId] = _.without(grainIdToSessionHashes[session.grainId], hashedSessionId);
                                                                       //
      if (grainIdToSessionHashes[session.grainId].length === 0) {      // 78
        grainIdToSessionHashes[session.grainId] = undefined;           // 79
        grainIdCount = grainIdCount - 1;                               // 80
        changed.activeGrains = grainIdCount;                           // 81
      }                                                                //
                                                                       //
      if (!_.isEmpty(changed)) {                                       // 84
        _this.changed("systemStatus", "globalStatus", changed);        // 85
      }                                                                //
    }                                                                  //
  });                                                                  //
  this.onStop(function () {                                            // 89
    handle.stop();                                                     // 90
  });                                                                  //
  this.ready();                                                        // 92
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=system-status-server.js.map
