(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/proxy.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require("crypto");                                    // 17
var ChildProcess = Npm.require("child_process");                       // 18
var Fs = Npm.require("fs");                                            // 19
var Path = Npm.require("path");                                        // 20
var Future = Npm.require("fibers/future");                             // 21
var Http = Npm.require("http");                                        // 22
var Url = Npm.require("url");                                          // 23
var Promise = Npm.require("es6-promise").Promise;                      // 24
var Capnp = Npm.require("capnp");                                      // 25
var Net = Npm.require("net");                                          // 26
                                                                       //
var ByteStream = Capnp.importSystem("sandstorm/util.capnp").ByteStream;
var ApiSession = Capnp.importSystem("sandstorm/api-session.capnp").ApiSession;
var WebSession = Capnp.importSystem("sandstorm/web-session.capnp").WebSession;
var HackSession = Capnp.importSystem("sandstorm/hack-session.capnp");  // 31
var Supervisor = Capnp.importSystem("sandstorm/supervisor.capnp").Supervisor;
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;   // 33
                                                                       //
var storeReferralProgramInfoApiTokenCreated = function (db, accountId, identityId, apiTokenAccountId) {
  // From the Referral program's perspective, if Bob's Account has no referredByComplete, then we
  // update Bob's Identity to say it's referredBy Alice's Account (which is apiTokenAccountId).
  check(accountId, String);                                            // 38
  check(identityId, String);                                           // 39
  check(apiTokenAccountId, String);                                    // 40
                                                                       //
  // Bail out early if quota enforcement is disabled.                  //
  if (!Meteor.settings["public"].quotaEnabled) {                       // 43
    return;                                                            // 44
  }                                                                    //
                                                                       //
  var aliceAccountId = apiTokenAccountId;                              // 47
  var bobAccountId = accountId;                                        // 48
  var bobIdentityId = identityId;                                      // 49
                                                                       //
  if (Meteor.users.find({                                              // 51
    _id: bobAccountId,                                                 // 52
    referredByComplete: { $exists: true }                              // 53
  }).count() > 0) {                                                    //
    return;                                                            // 55
  }                                                                    //
                                                                       //
  // Only actually update Bob's Identity ID if there is no referredBy.
  Meteor.users.update({ _id: bobIdentityId, referredBy: { $exists: false } }, { $set: { referredBy: aliceAccountId } });
};                                                                     //
                                                                       //
function referralProgramLogSharingTokenUse(db, bobAccountId) {         // 64
  // Hooray! The sharing token is valid! Someone (let's call them Charlie) is going to get a UiView
  // to this grain!  This means that the user who created this apiToken knows how to use the 'share
  // access' interface. Let's call them Bob.                           //
  //                                                                   //
  // If Bob's Account.referredByComplete is not yet set, then look at Bob's Identities and take the
  // first referredBy we find -- let's call that Alice.                //
  //                                                                   //
  // We copy Alice's account ID to Bob's Account.referredByComplete, and then update Alice's
  // referredIdentityIds to point at Bob's Identity, and then remove the referredBy from Bob's
  // Identity since it has become redundant.                           //
  //                                                                   //
  // Implementation note: this does mean that Alice can get referral credit for Bob by sharing a
  // link with Bob, even if Bob already had an account.                //
                                                                       //
  // Bail out early if quota support is not enabled.                   //
  if (!Meteor.settings["public"].quotaEnabled) {                       // 80
    return;                                                            // 81
  }                                                                    //
                                                                       //
  // Bail out if Bob has a referredByComplete.                         //
  if (Meteor.users.find({ _id: bobAccountId, referredByComplete: { $exists: true } }).count() > 0) {
    return;                                                            // 86
  }                                                                    //
                                                                       //
  // Look for a referredBy on any of Bob's identities.                 //
  var bobIdentityIds = SandstormDb.getUserIdentityIds(Meteor.users.findOne({ _id: bobAccountId }));
  var bobIdentityWithReferredBy = Meteor.users.findOne({               // 91
    _id: { $in: bobIdentityIds },                                      // 92
    referredBy: { $exists: true }                                      // 93
  });                                                                  //
                                                                       //
  if (!bobIdentityWithReferredBy) {                                    // 96
    return;                                                            // 97
  }                                                                    //
                                                                       //
  var aliceAccountId = bobIdentityWithReferredBy.referredBy;           // 100
                                                                       //
  // Store Bob's Account.referralCompletedBy.                          //
  var now = new Date();                                                // 103
  Meteor.users.update({                                                // 104
    _id: bobAccountId,                                                 // 105
    referredByComplete: { $exists: false }                             // 106
  }, {                                                                 //
    $set: {                                                            // 108
      referredByComplete: bobIdentityWithReferredBy.referredBy,        // 109
      referredCompleteDate: now                                        // 110
    }                                                                  //
  });                                                                  //
                                                                       //
  // Update Alice's Account.referredIdentityIds.                       //
  Meteor.users.update({ _id: aliceAccountId }, {                       // 115
    $push: { referredIdentityIds: bobIdentityWithReferredBy._id }      // 116
  });                                                                  //
                                                                       //
  // Remove now-redundant Bob identity referredBy.                     //
  Meteor.users.update({ _id: bobIdentityWithReferredBy._id }, {        // 120
    $unset: { referredBy: true }                                       // 121
  });                                                                  //
}                                                                      //
                                                                       //
// User-agent strings that should be allowed to use http basic authentication.
// These are regex matches, so ensure they are escaped properly with double
// backslashes. For security reasons, we MUST NOT whitelist any user-agents
// that may render html and execute embedded scripts.                  //
BASIC_AUTH_USER_AGENTS = ["git\\/", "GitHub-Hookshot\\/", "mirall\\/", "Mozilla\\/5\\.0 \\([^\\\\]*\\) mirall\\/", "Mozilla\\/5\\.0 \\(iOS\\) ownCloud-iOS\\/", "Mozilla\\/5\\.0 \\(Android\\) ownCloud-android\\/", "litmus\\/"];
BASIC_AUTH_USER_AGENTS_REGEX = new RegExp("^(" + BASIC_AUTH_USER_AGENTS.join("|") + ")", "");
                                                                       //
var SESSION_PROXY_TIMEOUT = 60000;                                     // 140
                                                                       //
var sandstormCoreFactory = makeSandstormCoreFactory();                 // 142
var backendAddress = "unix:" + (SANDSTORM_ALTHOME || "") + Backend.socketPath;
var sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);
var sandstormBackend = sandstormBackendConnection.restore(null, Backend);
                                                                       //
// TODO(cleanup): This initilization belongs with the rest of our package initialization in
//   db-deprecated.js. We can't put it there now because we need to contruct sandstormCoreFactory first.
globalBackend = new SandstormBackend(globalDb, sandstormBackend);      // 149
Meteor.onConnection(function (connection) {                            // 150
  connection.sandstormBackend = globalBackend;                         // 151
});                                                                    //
                                                                       //
// We've observed a problem in production where occasionally the front-end stops talking to the
// back-end. It happens very rarely -- like once a month -- and we've been unable to reproduce it
// in testing, making it very hard to debug. The problem appears both on Sandstorm and Blackrock.
// Restarting the node process (and only the node process) always fixes the problem.
//                                                                     //
// Here, I've added some code that attempts to detect the problem by doing a health check
// periodically and then remaking the connection if it seems broken. We'll see if this helps!
var backendHealthy = true;                                             // 161
var disconnectCount = 0;                                               // 162
Meteor.setInterval(function () {                                       // 163
  if (!backendHealthy) {                                               // 164
    if (disconnectCount++ > 2) process.abort();                        // 165
    console.error("error: Backend hasn't responded in 30 seconds! Reconnecting.");
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(true);
    sandstormBackendConnection.close();                                // 168
    sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);
    sandstormBackend = sandstormBackendConnection.restore(null, Backend);
    globalBackend._backendCap = sandstormBackend;                      // 171
  }                                                                    //
                                                                       //
  var debugLog = !backendHealthy;                                      // 174
  backendHealthy = false;                                              // 175
  var promise = sandstormBackend.ping().then(function () {             // 176
    backendHealthy = true;                                             // 177
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(false);
  }, function (err) {                                                  //
    console.error("error: Backend ping threw error!", err.stack);      // 180
    // The connection will be remade on the next interval. Note that we do NOT normally observe
    // exceptions being thrown for this problem; we see the connection simply stop responding.
    // So we don't expect this branch to execute in any case.          //
  });                                                                  //
                                                                       //
  if (debugLog) {                                                      // 186
    console.log("capnp.js: outer promise:", promise);                  // 187
  }                                                                    //
}, 30000);                                                             //
                                                                       //
// =======================================================================================
// Meteor context <-> Async Node.js context adapters                   //
// TODO(cleanup):  Move to a different file.                           //
                                                                       //
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {    // 195
  callback();                                                          // 196
});                                                                    //
                                                                       //
inMeteor = function (callback) {                                       // 199
  // Calls the callback in a Meteor context.  Returns a Promise for its result.
  return new Promise(function (resolve, reject) {                      // 201
    inMeteorInternal(function () {                                     // 202
      try {                                                            // 203
        resolve(callback());                                           // 204
      } catch (err) {                                                  //
        reject(err);                                                   // 206
      }                                                                //
    });                                                                //
  });                                                                  //
};                                                                     //
                                                                       //
promiseToFuture = function (promise) {                                 // 212
  var result = new Future();                                           // 213
  promise.then(result["return"].bind(result), result["throw"].bind(result));
  return result;                                                       // 215
};                                                                     //
                                                                       //
waitPromise = function (promise) {                                     // 218
  return promiseToFuture(promise).wait();                              // 219
};                                                                     //
                                                                       //
// =======================================================================================
// API for creating / starting grains from Meteor methods.             //
                                                                       //
var proxiesByHostId = {};                                              // 225
                                                                       //
Meteor.methods({                                                       // 227
  newGrain: function (packageId, command, title, identityId) {         // 228
    // Create and start a new grain.                                   //
                                                                       //
    check(packageId, String);                                          // 231
    check(command, Object); // Manifest.Command from package.capnp.    // 232
    check(title, String);                                              // 233
    check(identityId, String);                                         // 234
                                                                       //
    if (!this.userId) {                                                // 236
      throw new Meteor.Error(403, "Unauthorized", "Must be logged in to create grains.");
    }                                                                  //
                                                                       //
    if (!globalDb.userHasIdentity(this.userId, identityId)) {          // 240
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    if (!isSignedUpOrDemo()) {                                         // 244
      throw new Meteor.Error(403, "Unauthorized", "Only invited users or demo users can create grains.");
    }                                                                  //
                                                                       //
    if (isUserOverQuota(Meteor.user())) {                              // 249
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");
    }                                                                  //
                                                                       //
    var pkg = Packages.findOne(packageId);                             // 254
    var isDev = false;                                                 // 255
    if (!pkg) {                                                        // 256
      // Maybe they wanted a dev package.  Check there too.            //
      pkg = DevPackages.findOne(packageId);                            // 258
      isDev = true;                                                    // 259
    }                                                                  //
                                                                       //
    if (!pkg) {                                                        // 262
      throw new Meteor.Error(404, "Not Found", "No such package is installed.");
    }                                                                  //
                                                                       //
    var appId = pkg.appId;                                             // 266
    var manifest = pkg.manifest;                                       // 267
    var grainId = Random.id(22); // 128 bits of entropy                // 268
    Grains.insert({                                                    // 269
      _id: grainId,                                                    // 270
      packageId: packageId,                                            // 271
      appId: appId,                                                    // 272
      appVersion: manifest.appVersion,                                 // 273
      userId: this.userId,                                             // 274
      identityId: identityId,                                          // 275
      title: title,                                                    // 276
      "private": true                                                  // 277
    });                                                                //
                                                                       //
    globalBackend.startGrainInternal(packageId, grainId, this.userId, command, true, isDev);
    globalBackend.updateLastActive(grainId, this.userId, identityId);  // 281
    return grainId;                                                    // 282
  },                                                                   //
                                                                       //
  openSession: function (grainId, identityId, cachedSalt) {            // 285
    // Open a new UI session on an existing grain.  Starts the grain if it is not already
    // running.                                                        //
                                                                       //
    check(grainId, String);                                            // 289
    check(identityId, Match.OneOf(undefined, null, String));           // 290
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 291
                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    if (!Grains.findOne({ _id: grainId })) {                           // 297
      throw new Meteor.Error(404, "Grain not found", "Grain ID: " + grainId);
    }                                                                  //
                                                                       //
    var db = this.connection.sandstormDb;                              // 301
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 302
    if (!SandstormPermissions.mayOpenGrain(db, { grain: { _id: grainId, identityId: identityId } })) {
      throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
    }                                                                  //
                                                                       //
    var opened = globalBackend.openSessionInternal(grainId, this.userId, identityId, null, null, cachedSalt);
    var result = opened.methodResult;                                  // 310
    var proxy = new Proxy(grainId, this.userId, result.sessionId, result.hostId, result.tabId, identityId, false, opened.supervisor);
    proxiesByHostId[result.hostId] = proxy;                            // 313
    return result;                                                     // 314
  },                                                                   //
                                                                       //
  openSessionFromApiToken: function (params, identityId, cachedSalt) {
    // Given an API token, either opens a new WebSession to the underlying grain or returns a
    // path to which the client should redirect in order to open such a session.
                                                                       //
    check(params, {                                                    // 321
      token: String,                                                   // 322
      incognito: Boolean                                               // 323
    });                                                                //
    check(identityId, Match.OneOf(undefined, null, String));           // 325
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 326
                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    if (!identityId && globalDb.getOrganizationDisallowGuests()) {     // 332
      throw new Meteor.Error("guestDisallowed", "This Sandstorm server does not allow " + "guests or anonymous users");
    }                                                                  //
                                                                       //
    var token = params.token;                                          // 337
    var incognito = params.incognito;                                  // 338
    var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
    var apiToken = ApiTokens.findOne(hashedToken);                     // 340
    validateWebkey(apiToken);                                          // 341
    var grain = Grains.findOne({ _id: apiToken.grainId });             // 342
    if (!grain) {                                                      // 343
      throw new Meteor.Error(404, "Grain not found", "Grain ID: " + apiToken.grainId);
    }                                                                  //
                                                                       //
    if (apiToken.accountId) {                                          // 347
      referralProgramLogSharingTokenUse(globalDb, apiToken.accountId);
    }                                                                  //
                                                                       //
    var title = undefined;                                             // 351
    if (grain.userId === apiToken.accountId) {                         // 352
      title = grain.title;                                             // 353
    } else {                                                           //
      var sharerToken = apiToken.identityId && ApiTokens.findOne({     // 355
        grainId: apiToken.grainId,                                     // 357
        "owner.user.identityId": apiToken.identityId                   // 358
      }, {                                                             //
        sort: {                                                        // 360
          "owner.user.lastUsed": -1                                    // 361
        }                                                              //
      });                                                              //
      if (sharerToken) {                                               // 364
        title = sharerToken.owner.user.title;                          // 365
      } else {                                                         //
        title = "shared grain";                                        // 367
      }                                                                //
    }                                                                  //
                                                                       //
    if (this.userId && !incognito) {                                   // 371
      if (identityId != apiToken.identityId && identityId != grain.identityId && !ApiTokens.findOne({ "owner.user.identityId": identityId, parentToken: hashedToken })) {
        var owner = { user: { identityId: identityId, title: title } };
                                                                       //
        // Create a new API token for the identity redeeming this token.
        var result = SandstormPermissions.createNewApiToken(globalDb, { rawParentToken: token }, apiToken.grainId, apiToken.petname, { allAccess: null }, owner);
        globalDb.addContact(apiToken.accountId, identityId);           // 379
                                                                       //
        // If the parent API token is forSharing and it has an accountId, then the logged-in user (call
        // them Bob) is about to access a grain owned by someone (call them Alice) and save a reference
        // to it as a new ApiToken. (For share-by-link, this occurs when viewing the grain. For
        // share-by-identity, this happens immediately.)               //
        if (result.parentApiToken) {                                   // 385
          var parentApiToken = result.parentApiToken;                  // 386
          if (parentApiToken.forSharing && parentApiToken.accountId) {
            storeReferralProgramInfoApiTokenCreated(globalDb, this.userId, owner.user.identityId, parentApiToken.accountId);
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      return { redirectToGrain: apiToken.grainId };                    // 394
    } else {                                                           //
      if (!SandstormPermissions.mayOpenGrain(globalDb, { token: apiToken })) {
        throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
      }                                                                //
                                                                       //
      var opened = globalBackend.openSessionInternal(apiToken.grainId, null, null, title, apiToken, cachedSalt);
                                                                       //
      var result = opened.methodResult;                                // 404
      var proxy = new Proxy(apiToken.grainId, grain.userId, result.sessionId, result.hostId, result.tabId, identityId, false);
      proxy.apiToken = apiToken;                                       // 407
      proxiesByHostId[result.hostId] = proxy;                          // 408
      return result;                                                   // 409
    }                                                                  //
  },                                                                   //
                                                                       //
  keepSessionAlive: function (sessionId) {                             // 413
    // TODO(security):  Prevent draining someone else's quota by holding open several grains shared
    //   by them.                                                      //
    check(sessionId, String);                                          // 416
                                                                       //
    var session = Sessions.findAndModify({                             // 418
      query: { _id: sessionId },                                       // 419
      update: { $set: { timestamp: new Date().getTime() } },           // 420
      fields: { grainId: 1, identityId: 1, hostId: 1 }                 // 421
    });                                                                //
                                                                       //
    if (session) {                                                     // 424
      // Session still present in database, so send keep-alive to backend.
      try {                                                            // 426
        var grainId = session.grainId;                                 // 427
        var hostId = session.hostId;                                   // 428
        var supervisor = proxiesByHostId[hostId] && proxiesByHostId[hostId].supervisor;
        if (!supervisor) {                                             // 430
          supervisor = globalBackend.continueGrain(grainId).supervisor;
        }                                                              //
                                                                       //
        waitPromise(supervisor.keepAlive());                           // 434
        globalBackend.updateLastActive(grainId, this.userId, session.identityId);
      } catch (err) {                                                  //
        // Ignore disconnects, which imply that the grain shut down already. It'll start back up on
        // the next request, so whatever.                              //
        if (err.kjType !== "disconnected") {                           // 439
          throw err;                                                   // 440
        }                                                              //
      }                                                                //
                                                                       //
      return true;                                                     // 444
    } else {                                                           //
      return false;                                                    // 446
    }                                                                  //
  },                                                                   //
                                                                       //
  shutdownGrain: function (grainId) {                                  // 450
    check(grainId, String);                                            // 451
    var grain = Grains.findOne(grainId);                               // 452
    if (!grain || !this.userId || grain.userId !== this.userId) {      // 453
      throw new Meteor.Error(403, "Unauthorized", "User is not the owner of this grain");
    }                                                                  //
                                                                       //
    waitPromise(globalBackend.shutdownGrain(grainId, grain.userId, true));
  }                                                                    //
});                                                                    //
                                                                       //
var validateWebkey = function (apiToken, refreshedExpiration) {        // 461
  // Validates that `apiToken` is a valid UiView webkey, throwing an exception if it is not. If
  // `refreshedExpiration` is set and if the token has an `expiresIfUnused` field, then the
  // `expiresIfUnused` field is reset to `refreshedExpiration`.        //
                                                                       //
  if (!apiToken) {                                                     // 466
    throw new Meteor.Error(403, "Invalid authorization token");        // 467
  }                                                                    //
                                                                       //
  if (apiToken.revoked) {                                              // 470
    throw new Meteor.Error(403, "Authorization token has been revoked");
  }                                                                    //
                                                                       //
  if (apiToken.owner && !("webkey" in apiToken.owner)) {               // 474
    throw new Meteor.Error(403, "Unauthorized to open non-webkey token.");
  }                                                                    //
                                                                       //
  if (apiToken.expires && apiToken.expires.getTime() <= Date.now()) {  // 478
    throw new Meteor.Error(403, "Authorization token expired");        // 479
  }                                                                    //
                                                                       //
  if (apiToken.expiresIfUnused) {                                      // 482
    if (apiToken.expiresIfUnused.getTime() <= Date.now()) {            // 483
      throw new Meteor.Error(403, "Authorization token expired");      // 484
    } else if (refreshedExpiration) {                                  //
      ApiTokens.update(apiToken._id, { $set: { expiresIfUnused: refreshedExpiration } });
    } else {                                                           //
      // It's getting used now, so clear the expiresIfUnused field.    //
      ApiTokens.update(apiToken._id, { $set: { expiresIfUnused: null } });
    }                                                                  //
  }                                                                    //
                                                                       //
  if (apiToken.objectId || apiToken.frontendRef) {                     // 493
    throw new Meteor.Error(403, "ApiToken refers to a non-webview Capability.");
  }                                                                    //
};                                                                     //
                                                                       //
// Used by shared/grain.js (which sounds like a broken dependency, since this code is only available
// on the server)                                                      //
getGrainSize = function (supervisor, oldSize) {                        // 500
  var promise = undefined;                                             // 501
  if (oldSize === undefined) {                                         // 502
    promise = supervisor.getGrainSize();                               // 503
  } else {                                                             //
    promise = supervisor.getGrainSizeWhenDifferent(oldSize);           // 505
  }                                                                    //
                                                                       //
  var promise2 = promise.then(function (result) {                      // 508
    return parseInt(result.size);                                      // 508
  });                                                                  //
  promise2.cancel = function () {                                      // 509
    promise.cancel();                                                  // 509
  };                                                                   //
                                                                       //
  return promise2;                                                     // 511
};                                                                     //
                                                                       //
Meteor.startup(function () {                                           // 514
  var shutdownApp = function (appId) {                                 // 515
    Grains.find({ appId: appId }).forEach(function (grain) {           // 516
      waitPromise(globalBackend.shutdownGrain(grain._id, grain.userId));
    });                                                                //
  };                                                                   //
                                                                       //
  DevPackages.find().observe({                                         // 521
    removed: function (devPackage) {                                   // 522
      shutdownApp(devPackage.appId);                                   // 522
    },                                                                 //
                                                                       //
    changed: function (oldDevPackage, newDevPackage) {                 // 524
      shutdownApp(oldDevPackage.appId);                                // 525
      if (oldDevPackage.appId !== newDevPackage.appId) {               // 526
        shutdownApp(newDevPackage.appId);                              // 527
      }                                                                //
    },                                                                 //
                                                                       //
    added: function (devPackage) {                                     // 531
      shutdownApp(devPackage.appId);                                   // 531
    }                                                                  //
  });                                                                  //
                                                                       //
  Sessions.find().observe({                                            // 534
    removed: function (session) {                                      // 535
      var proxy = proxiesByHostId[session.hostId];                     // 536
      delete proxiesByHostId[session.hostId];                          // 537
      if (proxy) {                                                     // 538
        proxy.close();                                                 // 539
      }                                                                //
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
// Kill off sessions idle for >~3 minutes.                             //
var TIMEOUT_MS = 180000;                                               // 546
var gcSessions = function () {                                         // 547
  var now = new Date().getTime();                                      // 548
  Sessions.remove({ timestamp: { $lt: now - TIMEOUT_MS } });           // 549
};                                                                     //
                                                                       //
SandstormDb.periodicCleanup(TIMEOUT_MS, gcSessions);                   // 552
                                                                       //
var getProxyForHostId = function (hostId, isAlreadyOpened) {           // 554
  // Get the Proxy corresponding to the given grain session host, possibly (re)creating it if it
  // doesn't already exist. The first request on the session host will always create a new proxy.
  // Later requests may create a proxy if they go to a different front-end replica or if the
  // front-end was restarted.                                          //
  check(hostId, String);                                               // 559
                                                                       //
  return Promise.resolve(undefined).then(function () {                 // 561
    var proxy = proxiesByHostId[hostId];                               // 562
    if (proxy) {                                                       // 563
      return proxy;                                                    // 564
    } else {                                                           //
      // Set table entry to null for now so that we can detect if it is concurrently deleted.
      proxiesByHostId[hostId] = null;                                  // 567
                                                                       //
      return inMeteor(function () {                                    // 569
        var session = Sessions.findOne({ hostId: hostId });            // 570
        if (!session) {                                                // 571
          if (isAlreadyOpened) {                                       // 572
            return new Promise(function (resolve, reject) {            // 573
              var observer = undefined;                                // 574
              var task = Meteor.setTimeout(function () {               // 575
                observer.stop();                                       // 576
                reject(new Meteor.Error(504, "Requested session that no longer exists, and " + "timed out waiting for client to restore it. This can happen if you have " + "opened an app's content in a new window and then closed it in the " + "UI. If you see this error *inside* the Sandstorm UI, please report a " + "bug and describe the circumstances of the error."));
              }, SESSION_PROXY_TIMEOUT);                               //
              observer = Sessions.find({ hostId: hostId }).observe({   // 583
                added: function () {                                   // 584
                  observer.stop();                                     // 585
                  Meteor.clearTimeout(task);                           // 586
                  resolve(getProxyForHostId(hostId, false));           // 587
                }                                                      //
              });                                                      //
            });                                                        //
          } else {                                                     //
            // Does not appear to be a valid session host.             //
            return undefined;                                          // 593
          }                                                            //
        }                                                              //
                                                                       //
        var apiToken = undefined;                                      // 597
        if (session.hashedToken) {                                     // 598
          apiToken = ApiTokens.findOne({ _id: session.hashedToken });  // 599
          // We don't have to fully validate the API token here because if it changed the session
          // would have been deleted.                                  //
          if (!apiToken) {                                             // 602
            throw new Meteor.Error(410, "ApiToken has been deleted");  // 603
          }                                                            //
        }                                                              //
                                                                       //
        var grain = Grains.findOne(session.grainId);                   // 607
        if (!grain) {                                                  // 608
          // Grain was deleted, I guess.                               //
          throw new Meteor.Error(410, "Resource has been deleted");    // 610
        }                                                              //
                                                                       //
        // Note that we don't need to call mayOpenGrain() because the existence of a session
        // implies this check was already performed.                   //
                                                                       //
        var proxy = new Proxy(grain._id, grain.userId, session._id, hostId, session.tabId, session.identityId, false);
        if (apiToken) proxy.apiToken = apiToken;                       // 618
                                                                       //
        // Only add the proxy to the table if it was not concurrently deleted (which could happen
        // e.g. if the user's access was revoked).                     //
        if (hostId in proxiesByHostId) {                               // 622
          proxiesByHostId[hostId] = proxy;                             // 623
        } else {                                                       //
          throw new Meteor.Error(403, "Session was concurrently closed.");
        }                                                              //
                                                                       //
        return proxy;                                                  // 628
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
// =======================================================================================
// API tokens                                                          //
                                                                       //
var ApiSessionProxies = (function () {                                 //
  // Class that caches proxies for requests that come in through HTTP API endpoints. Such
  // requests are not associated with any entry in the `Sessions` collection, so we need
  // some other means of detecting unused proxies and closing them. This class works by keeping
  // the proxies in two buckets. When a proxy is used, it is put in the new bucket. Periodically,
  // the old bucket is emptied and the contents of new bucket are moved to the old bucket.
  //                                                                   //
  // Each bucket is a two-level map, keyed by hashes of the tokens and their `ApiSession.Params`
  // structs.                                                          //
                                                                       //
  function ApiSessionProxies(intervalMillis) {                         // 647
    var _this2 = this;                                                 //
                                                                       //
    babelHelpers.classCallCheck(this, ApiSessionProxies);              //
                                                                       //
    check(intervalMillis, Number);                                     // 648
    this.newBucket = {};                                               // 649
    this.oldBucket = {};                                               // 650
    this.interval = Meteor.setInterval(function () {                   // 651
      for (var oldHashedToken in babelHelpers.sanitizeForInObject(_this2.oldBucket)) {
        for (var oldHashedParams in babelHelpers.sanitizeForInObject(_this2.oldBucket[oldHashedToken])) {
          var newProxy = _this2.newBucket[oldHashedToken] && _this2.newBucket[oldHashedToken][oldHashedParams];
          var oldProxy = _this2.oldBucket[oldHashedToken][oldHashedParams];
          if (oldProxy && !newProxy) {                                 // 657
            if (Object.keys(oldProxy.websockets).length > 0) {         // 658
              // A client has an open websocket. Keep the proxy around.
              _this2.put(oldHashedToken, oldHashedParams, oldProxy);   // 660
            } else {                                                   //
              // We can close this proxy and forget about it.          //
              oldProxy.close();                                        // 663
            }                                                          //
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      _this2.oldBucket = _this2.newBucket;                             // 669
      _this2.newBucket = {};                                           // 670
    }, intervalMillis);                                                //
  }                                                                    //
                                                                       //
  ApiSessionProxies.prototype.get = (function () {                     // 637
    function get(hashedToken, hashedParams) {                          // 674
      if (this.newBucket[hashedToken] && this.newBucket[hashedToken][hashedParams]) {
        return this.newBucket[hashedToken][hashedParams];              // 676
      } else if (this.oldBucket[hashedToken] && this.oldBucket[hashedToken][hashedParams]) {
        var proxy = this.oldBucket[hashedToken][hashedParams];         // 678
        delete this.oldBucket[hashedToken][hashedParams];              // 679
        this.put(hashedToken, hashedParams, proxy);                    // 680
        return proxy;                                                  // 681
      } else {                                                         //
        return null;                                                   // 683
      }                                                                //
    }                                                                  //
                                                                       //
    return get;                                                        //
  })();                                                                //
                                                                       //
  ApiSessionProxies.prototype.put = (function () {                     // 637
    function put(hashedToken, hashedParams, proxy) {                   // 687
      if (!this.newBucket[hashedToken]) {                              // 688
        this.newBucket[hashedToken] = {};                              // 689
      }                                                                //
                                                                       //
      this.newBucket[hashedToken][hashedParams] = proxy;               // 692
    }                                                                  //
                                                                       //
    return put;                                                        //
  })();                                                                //
                                                                       //
  ApiSessionProxies.prototype.removeProxiesOfToken = (function () {    // 637
    function removeProxiesOfToken(hashedToken) {                       // 695
      if (hashedToken in this.oldBucket) {                             // 696
        for (var hashedParams in babelHelpers.sanitizeForInObject(this.oldBucket[hashedToken])) {
          this.oldBucket[hashedToken][hashedParams].close();           // 698
        }                                                              //
                                                                       //
        delete this.oldBucket[hashedToken];                            // 701
      }                                                                //
                                                                       //
      if (hashedToken in this.newBucket) {                             // 704
        for (var hashedParams in babelHelpers.sanitizeForInObject(this.newBucket[hashedToken])) {
          this.newBucket[hashedToken][hashedParams].close();           // 706
        }                                                              //
                                                                       //
        delete this.newBucket[hashedToken];                            // 709
      }                                                                //
    }                                                                  //
                                                                       //
    return removeProxiesOfToken;                                       //
  })();                                                                //
                                                                       //
  return ApiSessionProxies;                                            //
})();                                                                  //
                                                                       //
var apiSessionProxies = new ApiSessionProxies(3 * 60 * 1000);          // 714
                                                                       //
Meteor.startup(function () {                                           // 716
  var clearSessionsAndProxies = function (token) {                     // 717
    // Clears all sessions and API proxies associated with `token` or any token that is downstream
    // in the sharing graph.                                           //
    // TODO(soon): Only clear sessions and proxies for which the permissions have changed.
    var downstream = SandstormPermissions.downstreamTokens(globalDb, { token: token });
    downstream.push(token);                                            // 722
    var identityIds = [];                                              // 723
    var tokenIds = [];                                                 // 724
                                                                       //
    downstream.forEach(function (token) {                              // 726
      apiSessionProxies.removeProxiesOfToken(token._id);               // 727
                                                                       //
      tokenIds.push(token._id);                                        // 729
      if (token.owner && token.owner.user) {                           // 730
        identityIds.push(token.owner.user.identityId);                 // 731
      }                                                                //
    });                                                                //
                                                                       //
    Sessions.find({                                                    // 735
      grainId: token.grainId,                                          // 736
      $or: [{ identityId: { $in: identityIds } }, { hashedToken: { $in: tokenIds } }]
    }, {                                                               //
      fields: { hostId: 1 }                                            // 741
    }).forEach(function (session) {                                    //
      var proxy = proxiesByHostId[session.hostId];                     // 743
      if (proxy) {                                                     // 744
        proxy.close();                                                 // 745
      }                                                                //
                                                                       //
      delete proxiesByHostId[session.hostId];                          // 748
    });                                                                //
                                                                       //
    Sessions.remove({                                                  // 751
      grainId: token.grainId,                                          // 752
      $or: [{ identityId: { $in: identityIds } }, { hashedToken: { $in: tokenIds } }]
    });                                                                //
  };                                                                   //
                                                                       //
  Grains.find().observe({                                              // 760
    changed: function (newGrain, oldGrain) {                           // 761
      if (oldGrain["private"] != newGrain["private"]) {                // 762
        Sessions.remove({ grainId: oldGrain._id, identityId: { $ne: oldGrain.identityId } });
        ApiTokens.find({ grainId: oldGrain._id }).forEach(function (apiToken) {
          apiSessionProxies.removeProxiesOfToken(apiToken._id);        // 765
        });                                                            //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  ApiTokens.find({ grainId: { $exists: true }, objectId: { $exists: false } }).observe({
    added: function (newApiToken) {                                    // 772
      // TODO(soon): Unfortunately, added() gets called for all existing role assignments when the
      //   front-end restarts, meaning clearing sessions here will cause people's views to refresh
      //   on server upgrade, which is not a nice user experience. It's also sad to force-refresh
      //   people when they gained new permissions since they might be in the middle of something,
      //   and it's not strictly necessary for security. OTOH, it's sad to be non-reactive. Maybe
      //   we should notify people that they have new permissions and let them click a thing to
      //   refresh?                                                    //
      //clearSessions(roleAssignment.grainId, roleAssignment.recipient);
      //clearApiProxies(roleAssignment.grainId);                       //
    },                                                                 //
                                                                       //
    changed: function (newApiToken, oldApiToken) {                     // 784
      if (!_.isEqual(newApiToken.roleAssignment, oldApiToken.roleAssignment) || !_.isEqual(newApiToken.revoked, oldApiToken.revoked)) {
        clearSessionsAndProxies(newApiToken);                          // 787
      }                                                                //
    },                                                                 //
                                                                       //
    removed: function (oldApiToken) {                                  // 791
      clearSessionsAndProxies(oldApiToken);                            // 792
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
function getApiSessionParams(request) {                                // 797
  var params = {};                                                     // 798
  if ("x-sandstorm-passthrough" in request.headers) {                  // 799
    var optIns = request.headers["x-sandstorm-passthrough"].split(",").map(function (s) {
      return s.trim();                                                 // 802
    });                                                                //
    // The only currently supported passthrough value is 'address', but others could be useful in
    // the future.                                                     //
                                                                       //
    if (optIns.indexOf("address") !== -1) {                            // 806
      // Sadly, we can't use request.socket.remoteFamily because it's not available in the
      // (rather-old) version of node that comes in the Meteor bundle we're using. Hence this
      // hackery.                                                      //
      var addressToPass = request.socket.remoteAddress;                // 810
      if (isRfc1918OrLocal(addressToPass) && "x-real-ip" in request.headers) {
        // Allow overriding the socket's remote address with X-Real-IP header if the request comes
        // from either localhost or an RFC1918 address. These are not useful for geolocation anyway.
        addressToPass = request.headers["x-real-ip"];                  // 814
      }                                                                //
                                                                       //
      if (Net.isIPv4(addressToPass)) {                                 // 817
        // Map IPv4 addresses in IPv6.                                 //
        // This conveniently comes out to a 48-bit number, which is precisely representable in a
        // double (which has 53 mantissa bits). Thus we can avoid using Bignum/strings, which we
        // might otherwise need to precisely represent 64-bit fields.  //
        var v4Int = 0xFFFF00000000 + addressToPass.split(".").map(function (x) {
          return parseInt(x, 10);                                      // 823
        }).reduce(function (a, b) {                                    //
          return 256 * a + b;                                          // 824
        });                                                            //
        params.remoteAddress = {                                       // 825
          lower64: v4Int,                                              // 826
          upper64: 0                                                   // 827
        };                                                             //
      } else if (Net.isIPv6(addressToPass)) {                          //
        // TODO(test): Unit test this                                  //
        // Parse a valid v6 address.                                   //
        // Split into groups, then insert an appropriate number of 0's if :: was used.
        var groups = addressToPass.split(":");                         // 833
                                                                       //
        // Strip extra empty group in the case of a leading or trailing '::'.
        if (groups[0] === "") {                                        // 836
          groups.shift();                                              // 837
        }                                                              //
                                                                       //
        if (groups[groups.length - 1] === "") {                        // 840
          groups.pop();                                                // 841
        }                                                              //
                                                                       //
        var lastGroup = groups[groups.length - 1];                     // 844
        // Handle IPv4-mapped IPv6 addresses.  These end in a dotted-quad IPv4 address, which we
        // should expand into two groups of 4-character hex strings, like the rest of the address.
        if (Net.isIPv4(lastGroup)) {                                   // 847
          groups.pop();                                                // 848
          var _quad = lastGroup.split(".").map(function (x) {          // 849
            return parseInt(x, 10);                                    // 849
          });                                                          //
          groups.push((_quad[0] * 256 + _quad[1]).toString(16));       // 850
          groups.push((_quad[2] * 256 + _quad[3]).toString(16));       // 851
        }                                                              //
                                                                       //
        var groupsToAdd = 8 - groups.length;                           // 854
        var emptyGroupIndex = groups.indexOf("");                      // 855
        var cleanGroups = undefined;                                   // 856
        if (emptyGroupIndex !== -1) {                                  // 857
          var head = groups.slice(0, emptyGroupIndex);                 // 858
          // groupsToAdd + 1 because we sliced out the empty element   //
          var mid = Array(groupsToAdd + 1);                            // 860
          for (var i = 0; i < groupsToAdd + 1; i++) {                  // 861
            mid[i] = "0";                                              // 862
          }                                                            //
                                                                       //
          var tail = groups.slice(emptyGroupIndex + 1, groups.length);
          cleanGroups = [].concat(head, mid, tail);                    // 866
        } else {                                                       //
          cleanGroups = groups;                                        // 868
        }                                                              //
                                                                       //
        var ints = cleanGroups.map(function (x) {                      // 871
          return parseInt(x, 16);                                      // 871
        });                                                            //
        // We use strings because we'd lose data from loss of precision casting the 64-bit uints
        // into 53-bit-mantissa doubles.                               //
        params.remoteAddress = {                                       // 874
          upper64: quadToIntString(ints.slice(0, 4)),                  // 875
          lower64: quadToIntString(ints.slice(4, 8))                   // 876
        };                                                             //
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  return Capnp.serialize(ApiSession.Params, params);                   // 882
}                                                                      //
                                                                       //
// Used by server/drivers/external-ui/view.js                          //
getProxyForApiToken = function (token, request) {                      // 886
  check(token, String);                                                // 887
  var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
  var tabId = Crypto.createHash("sha256").update("tab:").update(hashedToken).digest("hex").slice(0, 32);
  var serializedParams = getApiSessionParams(request);                 // 891
  var hashedParams = Crypto.createHash("sha256").update(serializedParams).digest("base64");
  return Promise.resolve(undefined).then(function () {                 // 893
    var proxy = apiSessionProxies.get(hashedToken, hashedParams);      // 894
    if (proxy) {                                                       // 895
      if (proxy.expires && proxy.expires.getTime() <= Date.now()) {    // 896
        throw new Meteor.Error(403, "Authorization token expired");    // 897
      }                                                                //
                                                                       //
      return proxy;                                                    // 900
    } else {                                                           //
      return inMeteor(function () {                                    // 902
        var tokenInfo = ApiTokens.findOne(hashedToken);                // 903
        validateWebkey(tokenInfo);                                     // 904
                                                                       //
        var grain = Grains.findOne(tokenInfo.grainId);                 // 906
        if (!grain) {                                                  // 907
          // Grain was deleted, I guess.                               //
          throw new Meteor.Error(410, "Resource has been deleted");    // 909
        }                                                              //
                                                                       //
        var proxy = undefined;                                         // 912
        if (tokenInfo.userInfo) {                                      // 913
          throw new Error("API tokens created with arbitrary userInfo no longer supported");
        } else {                                                       //
          var identityId = null;                                       // 916
          if (tokenInfo.identityId && !tokenInfo.forSharing) {         // 917
            identityId = tokenInfo.identityId;                         // 918
          }                                                            //
                                                                       //
          proxy = new Proxy(tokenInfo.grainId, grain.userId, null, null, tabId, identityId, true);
          proxy.apiToken = tokenInfo;                                  // 922
          proxy.apiSessionParams = serializedParams;                   // 923
        }                                                              //
                                                                       //
        if (!SandstormPermissions.mayOpenGrain(globalDb, { token: tokenInfo })) {
          // Note that only public grains may be opened without a user ID.
          throw new Meteor.Error(403, "Unauthorized.");                // 928
        }                                                              //
                                                                       //
        if (tokenInfo.expires) {                                       // 931
          proxy.expires = tokenInfo.expires;                           // 932
        }                                                              //
                                                                       //
        apiSessionProxies.put(hashedToken, hashedParams, proxy);       // 935
                                                                       //
        return proxy;                                                  // 937
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
var apiUseBasicAuth = function (req, hostId) {                         // 943
  // If the request was to a token-specific host *and* the request has no Origin header (meaning
  // it is not an XHR from a web site), then we permit the use of HTTP basic auth. The reason we
  // prohibit XHRs is because we want to discourage people from using basic auth in a browser,
  // because the browser will cache the credentials, which exposes the hostname to XSRF attack
  // (which is mostly, but not completely, mitigated by the unguessable hostname). The reason we
  // want to allow basic auth at all is because a lot of existing client apps support only basic
  // auth. New clients and web-based clients should use `Authorization: bearer <token>` instead.
  if (globalDb.isTokenSpecificHostId(hostId) && !req.headers.origin) {
    return true;                                                       // 952
  }                                                                    //
                                                                       //
  // Historically, all API tokens were served from the same host rather than have token-specific
  // hosts. In this model, basic auth in browsers was far more dangerous since the hostname was
  // not a secret -- in fact, even benign attempts to use two APIs on the same server from the same
  // browser could interfere if using basic auth. Hence, basic auth was prohibited except when
  // coming from certain whitelisted clients known not to be browsers, e.g. Mirall, the ownCloud
  // client app.                                                       //
  //                                                                   //
  // Since many clients in the wild have already been configured to use the shared API host, we
  // must continue to support them, so this logic remains.             //
  var agent = req.headers["user-agent"];                               // 964
  return agent.match(BASIC_AUTH_USER_AGENTS_REGEX);                    // 965
};                                                                     //
                                                                       //
var apiTokenForRequest = function (req, hostId) {                      // 968
  // Extract the API token from the request.                           //
                                                                       //
  var auth = req.headers.authorization;                                // 971
  var token = undefined;                                               // 972
  if (auth && auth.slice(0, 7).toLowerCase() === "bearer ") {          // 973
    token = auth.slice(7).trim();                                      // 974
  } else if (auth && auth.slice(0, 6).toLowerCase() === "basic " && apiUseBasicAuth(req, hostId)) {
    token = new Buffer(auth.slice(6).trim(), "base64").toString().split(":")[1];
  } else {                                                             //
    token = undefined;                                                 // 979
  }                                                                    //
                                                                       //
  if (token && hostId !== "api") {                                     // 982
    // Verify that the token matches the specific host.                //
    if (hostId !== globalDb.apiHostIdForToken(token)) {                // 984
      token = undefined;                                               // 985
    }                                                                  //
  }                                                                    //
                                                                       //
  return token;                                                        // 989
};                                                                     //
                                                                       //
// =======================================================================================
// Routing to proxies.                                                 //
//                                                                     //
                                                                       //
// pre-meteor.js calls this                                            //
tryProxyUpgrade = function (hostId, req, socket, head) {               // 997
  // Attempt to handle a WebSocket upgrade by dispatching it to a grain. Returns a promise that
  // resolves true if an appropriate grain is found, false if there was no match (but the caller
  // should consider other host types, like static web publishing), or throws an error if the
  // request is definitely invalid.                                    //
                                                                       //
  if (globalDb.isApiHostId(hostId)) {                                  // 1003
    var token = apiTokenForRequest(req, hostId);                       // 1004
    if (token) {                                                       // 1005
      return getProxyForApiToken(token).then(function (proxy) {        // 1006
        // Meteor sets the timeout to five seconds. Change that back to two
        // minutes, which is the default value.                        //
        socket.setTimeout(120000);                                     // 1009
                                                                       //
        proxy.upgradeHandler(req, socket, head);                       // 1011
        return true;                                                   // 1012
      });                                                              //
    } else {                                                           //
      return Promise.resolve(false);                                   // 1015
    }                                                                  //
  } else {                                                             //
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {
      if (proxy) {                                                     // 1020
        // Cross-origin requests are not allowed on UI session hosts.  //
        var origin = req.headers.origin;                               // 1022
        if (origin !== PROTOCOL + "//" + req.headers.host) {           // 1023
          throw new Meteor.Error(403, "Detected illegal cross-origin WebSocket from: " + origin);
        }                                                              //
                                                                       //
        // Meteor sets the timeout to five seconds. Change that back to two
        // minutes, which is the default value.                        //
        socket.setTimeout(120000);                                     // 1029
                                                                       //
        proxy.upgradeHandler(req, socket, head);                       // 1031
        return true;                                                   // 1032
      } else {                                                         //
        return false;                                                  // 1034
      }                                                                //
    });                                                                //
  }                                                                    //
};                                                                     //
                                                                       //
// pre-meteor.js calls this                                            //
tryProxyRequest = function (hostId, req, res) {                        // 1041
  // Attempt to handle an HTTP request by dispatching it to a grain. Returns a promise that
  // resolves true if an appropriate grain is found, false if there was no match (but the caller
  // should consider other host types, like static web publishing), or throws an error if the
  // request is definitely invalid.                                    //
                                                                       //
  var hostIdHash = globalDb.isApiHostId(hostId);                       // 1047
  if (hostIdHash) {                                                    // 1048
    var _ret = (function () {                                          //
      // This is a request for the API host.                           //
                                                                       //
      res.setHeader("Access-Control-Allow-Origin", "*");               // 1051
                                                                       //
      if (req.method === "OPTIONS") {                                  // 1053
        // Reply to CORS preflight request.                            //
                                                                       //
        // All we want to do is permit APIs to be accessed from arbitrary origins. Since clients
        // must send a valid Authorization header, and since cookies are not used for
        // authorization, this is perfectly safe. In a sane world, we would only need to send back
        // 'Access-Control-Allow-Origin: *' and be done with it.       //
        //                                                             //
        // However, CORS demands that we explicitly whitelist individual methods and headers for
        // use cross-origin, as if this is somehow useful for implementing any practical security
        // policy (it isn't). To make matters worse, we are REQUIRED to enumerate each one
        // individually. We cannot just write '*' for these lists. WTF, CORS?
        //                                                             //
        // Luckily, the request tells us exactly what method and headers are being requested, so we
        // only need to copy those over, rather than create an exhaustive list. But this is still
        // overly complicated.                                         //
                                                                       //
        var accessControlHeaders = {                                   // 1070
          "Access-Control-Allow-Methods": "GET, HEAD, POST, PUT, PATCH, DELETE",
          "Access-Control-Max-Age": "3600"                             // 1072
        };                                                             //
                                                                       //
        // Copy all requested headers to the allowed headers list.     //
        var requestedHeaders = req.headers["access-control-request-headers"];
        if (requestedHeaders) {                                        // 1077
          accessControlHeaders["Access-Control-Allow-Headers"] = requestedHeaders;
        }                                                              //
                                                                       //
        // Add the requested method to the allowed methods list, if it's not there already.
        var requestedMethod = req.headers["access-control-request-method"];
        if (requestedMethod && !_.contains(["GET", "HEAD", "POST", "PUT", "PATCH", "DELETE"], requestedMethod)) {
          accessControlHeaders["Access-Control-Allow-Methods"] += ", " + requestedMethod;
        }                                                              //
                                                                       //
        for (header in babelHelpers.sanitizeForInObject(accessControlHeaders)) {
          res.setHeader(header, accessControlHeaders[header]);         // 1089
        }                                                              //
      }                                                                //
                                                                       //
      var errorHandler = function (err) {                              // 1093
        if (err instanceof Meteor.Error) {                             // 1094
          console.log("error: " + err);                                // 1095
          res.writeHead(err.error, err.reason, { "Content-Type": "text/plain" });
        } else {                                                       //
          res.writeHead(500, "Internal Server Error", { "Content-Type": "text/plain" });
        }                                                              //
                                                                       //
        res.end(err.stack);                                            // 1101
      };                                                               //
                                                                       //
      var token = apiTokenForRequest(req, hostId);                     // 1104
      if (token && req.headers["x-sandstorm-token-keepalive"]) {       // 1105
        inMeteor(function () {                                         // 1106
          var keepaliveDuration = parseInt(req.headers["x-sandstorm-token-keepalive"]);
          check(keepaliveDuration, Match.Integer);                     // 1108
          var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
          validateWebkey(ApiTokens.findOne(hashedToken), new Date(Date.now() + keepaliveDuration));
        }).then(function () {                                          //
          res.writeHead(200, {});                                      // 1112
          res.end();                                                   // 1113
        }, errorHandler);                                              //
      } else if (token) {                                              //
        getProxyForApiToken(token, req).then(function (proxy) {        // 1116
          proxy.requestHandler(req, res);                              // 1117
        }, errorHandler);                                              //
      } else {                                                         //
        // No token. Look up static API host info.                     //
        inMeteor(function () {                                         // 1121
          var apiHost = globalDb.collections.apiHosts.findOne(hostIdHash);
                                                                       //
          if (req.method === "OPTIONS") {                              // 1124
            // OPTIONS request with no authorization token.            //
                                                                       //
            var dav = ((apiHost || {}).options || {}).dav || [];       // 1127
            if (dav.length > 0) {                                      // 1128
              res.setHeader("DAV", dav.join(", "));                    // 1129
              res.setHeader("Access-Control-Expose-Headers", "DAV");   // 1130
            }                                                          //
                                                                       //
            res.writeHead(200, {});                                    // 1133
            res.end();                                                 // 1134
          } else {                                                     //
            var resources = (apiHost || {}).resources || {};           // 1136
            var path = SandstormDb.escapeMongoKey(req.url.split("?")[0]);
            if (path in resources) {                                   // 1138
              // Serve a static resource.                              //
              var resource = resources[path];                          // 1140
              console.log(path, resources, resource);                  // 1141
              if (resource.language) res.setHeader("Content-Language", resource.language);
              if (resource.encoding) res.setHeader("Content-Encoding", resource.encoding);
              res.writeHead(200, {                                     // 1144
                "Content-Type": resource.type                          // 1145
              });                                                      //
              res.end(resource.body);                                  // 1147
            } else {                                                   //
              if (apiUseBasicAuth(req, hostId)) {                      // 1149
                res.writeHead(401, {                                   // 1150
                  "Content-Type": "text/plain",                        // 1151
                  "WWW-Authenticate": "Basic realm='Sandstorm API'"    // 1152
                });                                                    //
              } else {                                                 //
                // TODO(someday): Display some sort of nifty API browser.
                res.writeHead(403, { "Content-Type": "text/plain" });  // 1156
              }                                                        //
                                                                       //
              res.end("Missing or invalid authorization header.\n\n" + "This address serves APIs, which allow external apps (such as a phone app) to\n" + "access data on your Sandstorm server. This address is not meant to be opened\n" + "in a regular browser.");
            }                                                          //
          }                                                            //
        }).then(function () {}, errorHandler);                         //
      }                                                                //
                                                                       //
      return {                                                         // 1168
        v: Promise.resolve(true)                                       //
      };                                                               //
    })();                                                              //
                                                                       //
    if (typeof _ret === "object") return _ret.v;                       //
  } else {                                                             //
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {
      if (proxy) {                                                     // 1172
        proxy.requestHandler(req, res);                                // 1173
        return true;                                                   // 1174
      } else {                                                         //
        return false;                                                  // 1176
      }                                                                //
    });                                                                //
  }                                                                    //
};                                                                     //
                                                                       //
// =======================================================================================
// Proxy class                                                         //
//                                                                     //
// Connects to a grain and exports it on a wildcard host.              //
//                                                                     //
                                                                       //
var Proxy = (function () {                                             //
  function Proxy(grainId, ownerId, sessionId, hostId, tabId, identityId, isApi, supervisor) {
    var _this3 = this;                                                 //
                                                                       //
    babelHelpers.classCallCheck(this, Proxy);                          //
                                                                       //
    this.grainId = grainId;                                            // 1190
    this.ownerId = ownerId;                                            // 1191
    this.identityId = identityId;                                      // 1192
    this.supervisor = supervisor; // note: optional parameter; we can reconnect
    this.sessionId = sessionId;                                        // 1194
    this.tabId = tabId;                                                // 1195
    this.isApi = isApi;                                                // 1196
    this.hasLoaded = false;                                            // 1197
    this.websockets = {};                                              // 1198
    this.websocketCounter = 0; // Used for generating unique socket IDs.
    if (sessionId) {                                                   // 1200
      if (!hostId) throw new Error("sessionId must come with hostId");
      if (isApi) throw new Error("API proxy shouldn't have sessionId");
      this.hostId = hostId;                                            // 1203
    } else {                                                           //
      if (!isApi) throw new Error("non-API proxy requires sessionId");
      if (hostId) throw new Error("API proxy sholudn't have hostId");  // 1206
    }                                                                  //
                                                                       //
    if (this.identityId) {                                             // 1209
      var identity = globalDb.getIdentity(this.identityId);            // 1210
      if (!identity) {                                                 // 1211
        throw new Error("identity not found: " + this.identityId);     // 1212
      }                                                                //
                                                                       //
      this.userInfo = {                                                // 1215
        displayName: { defaultText: identity.profile.name },           // 1216
        preferredHandle: identity.profile.handle,                      // 1217
        identityId: new Buffer(identity._id, "hex")                    // 1218
      };                                                               //
      if (identity.profile.pictureUrl) this.userInfo.pictureUrl = identity.profile.pictureUrl;
      if (identity.profile.pronoun) this.userInfo.pronouns = identity.profile.pronoun;
    } else {                                                           //
      this.userInfo = {                                                // 1223
        displayName: { defaultText: "Anonymous User" },                // 1224
        preferredHandle: "anonymous"                                   // 1225
      };                                                               //
    }                                                                  //
                                                                       //
    var _this = this;                                                  // 1229
                                                                       //
    this.requestHandler = function (request, response) {               // 1231
      if (_this3.sessionId) {                                          // 1232
        // Implement /_sandstorm-init for setting the session cookie.  //
        var url = Url.parse(request.url, true);                        // 1234
        if (url.pathname === "/_sandstorm-init" && url.query.sessionid === _this.sessionId) {
          _this.doSessionInit(request, response, url.query.path);      // 1236
          return;                                                      // 1237
        }                                                              //
      }                                                                //
                                                                       //
      Promise.resolve(undefined).then(function () {                    // 1241
        var contentLength = request.headers["content-length"];         // 1242
        if ((request.method === "POST" || request.method === "PUT") && (contentLength === undefined || contentLength > 1024 * 1024)) {
          // The input is either very long, or we don't know how long it is, so use streaming mode.
          return _this.handleRequestStreaming(request, response, contentLength, 0);
        } else {                                                       //
          return readAll(request).then(function (data) {               // 1248
            return _this.handleRequest(request, data, response, 0);    // 1249
          });                                                          //
        }                                                              //
      })["catch"](function (err) {                                     //
        _this.setHasLoaded();                                          // 1253
                                                                       //
        var body = err.stack;                                          // 1255
        if (err.cppFile) {                                             // 1256
          body += "\nC++ location:" + err.cppFile + ":" + (err.line || "??");
        }                                                              //
                                                                       //
        if (err.kjType) {                                              // 1260
          body += "\ntype: " + err.kjType;                             // 1261
        }                                                              //
                                                                       //
        if (response.headersSent) {                                    // 1264
          // Unfortunately, it's too late to tell the client what happened.
          console.error("HTTP request failed after response already sent:", body);
          response.end();                                              // 1267
        } else {                                                       //
          if (err instanceof Meteor.Error) {                           // 1269
            response.writeHead(err.error, err.reason, { "Content-Type": "text/plain" });
          } else {                                                     //
            response.writeHead(500, "Internal Server Error", { "Content-Type": "text/plain" });
          }                                                            //
                                                                       //
          response.end(body);                                          // 1275
        }                                                              //
      });                                                              //
    };                                                                 //
                                                                       //
    this.upgradeHandler = function (request, socket, head) {           // 1280
      _this.handleWebSocket(request, socket, head, 0)["catch"](function (err) {
        console.error("WebSocket setup failed:", err.stack);           // 1282
        // TODO(cleanup):  Manually send back a 500 response?          //
        socket.destroy();                                              // 1284
      });                                                              //
    };                                                                 //
  }                                                                    //
                                                                       //
  Proxy.prototype.close = (function () {                               // 1188
    function close() {                                                 // 1289
      for (var socketIdx in babelHelpers.sanitizeForInObject(this.websockets)) {
        this.websockets[socketIdx].destroy();                          // 1291
      };                                                               //
                                                                       //
      this.websockets = {};                                            // 1294
                                                                       //
      if (this.session) {                                              // 1296
        this.session.close();                                          // 1297
        delete this.session;                                           // 1298
      }                                                                //
                                                                       //
      if (this.uiView) {                                               // 1301
        this.uiView.close();                                           // 1302
        delete this.uiView;                                            // 1303
      }                                                                //
                                                                       //
      if (this.supervisor) {                                           // 1306
        this.supervisor.close();                                       // 1307
        delete this.supervisor;                                        // 1308
      }                                                                //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  Proxy.prototype.getConnection = (function () {                       // 1188
    function getConnection() {                                         // 1312
      if (!this.supervisor) {                                          // 1313
        this.supervisor = globalBackend.cap().getGrain(this.ownerId, this.grainId).supervisor;
        this.uiView = null;                                            // 1315
      }                                                                //
                                                                       //
      if (!this.uiView) {                                              // 1318
        this.uiView = this.supervisor.getMainView().view;              // 1319
      }                                                                //
    }                                                                  //
                                                                       //
    return getConnection;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewWebSession = (function () {                  // 1188
    function _callNewWebSession(request, userInfo) {                   // 1323
      var params = Capnp.serialize(WebSession.Params, {                // 1324
        basePath: PROTOCOL + "//" + request.headers.host,              // 1325
        userAgent: "user-agent" in request.headers ? request.headers["user-agent"] : "UnknownAgent/0.0",
        acceptableLanguages: "accept-language" in request.headers ? request.headers["accept-language"].split(",").map(function (s) {
          return s.trim();                                             // 1330
        }) : ["en-US", "en"]                                           //
      });                                                              //
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId, this.tabId), WebSession.typeId, params, new Buffer(this.tabId, "hex")).session;
    }                                                                  //
                                                                       //
    return _callNewWebSession;                                         //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewApiSession = (function () {                  // 1188
    function _callNewApiSession(request, userInfo) {                   // 1338
      var _this4 = this;                                               //
                                                                       //
      var serializedParams = this.apiSessionParams;                    // 1339
      if (!serializedParams) {                                         // 1340
        throw new Meteor.Error(500, "Should have already computed apiSessionParams.");
      }                                                                //
                                                                       //
      // TODO(someday): We are currently falling back to WebSession if we get any kind of error upon
      // calling newSession with an ApiSession._id.                    //
      // Eventually we'll remove this logic once we're sure apps have updated.
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId, this.tabId), ApiSession.typeId, serializedParams, new Buffer(this.tabId, "hex")).then(function (session) {
        return session.session;                                        // 1351
      }, function (err) {                                              //
        return _this4._callNewWebSession(request, userInfo);           // 1353
      });                                                              //
    }                                                                  //
                                                                       //
    return _callNewApiSession;                                         //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewSession = (function () {                     // 1188
    function _callNewSession(request, viewInfo) {                      // 1357
      var userInfo = _.clone(this.userInfo);                           // 1358
      var _this = this;                                                // 1359
      var promise = inMeteor(function () {                             // 1360
        var vertex = undefined;                                        // 1361
        if (_this.apiToken) {                                          // 1362
          vertex = { token: _this.apiToken };                          // 1363
        } else {                                                       //
          // (_this.identityId might be null; this is fine)            //
          vertex = { grain: { _id: _this.grainId, identityId: _this.identityId } };
        }                                                              //
                                                                       //
        var permissions = SandstormPermissions.grainPermissions(globalDb, vertex, viewInfo);
        if (!permissions) {                                            // 1370
          throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
        }                                                              //
                                                                       //
        Sessions.update({                                              // 1374
          _id: _this.sessionId                                         // 1375
        }, {                                                           //
          $set: {                                                      // 1377
            viewInfo: viewInfo,                                        // 1378
            permissions: permissions                                   // 1379
          }                                                            //
        });                                                            //
                                                                       //
        return permissions;                                            // 1383
      });                                                              //
                                                                       //
      return promise.then(function (permissions) {                     // 1386
        userInfo.permissions = permissions;                            // 1387
                                                                       //
        var numBytes = Math.ceil(permissions.length / 8);              // 1389
        var buf = new Buffer(numBytes);                                // 1390
        for (var ii = 0; ii < numBytes; ++ii) {                        // 1391
          buf[ii] = 0;                                                 // 1392
        }                                                              //
                                                                       //
        for (var ii = 0; ii < permissions.length; ++ii) {              // 1395
          var byteNum = Math.floor(ii / 8);                            // 1396
          var bitNum = ii % 8;                                         // 1397
          if (permissions[ii]) {                                       // 1398
            buf[byteNum] = buf[byteNum] | 1 << bitNum;                 // 1399
          }                                                            //
        }                                                              //
                                                                       //
        userInfo.deprecatedPermissionsBlob = buf;                      // 1403
                                                                       //
        if (_this.isApi) {                                             // 1405
          return _this._callNewApiSession(request, userInfo);          // 1406
        } else {                                                       //
          return _this._callNewWebSession(request, userInfo);          // 1408
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return _callNewSession;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.getSession = (function () {                          // 1188
    function getSession(request) {                                     // 1413
      var _this5 = this;                                               //
                                                                       //
      if (!this.session) {                                             // 1414
        (function () {                                                 //
          _this5.getConnection(); // make sure we're connected         // 1415
          var _this = _this5;                                          // 1416
          var promise = _this5.uiView.getViewInfo().then(function (viewInfo) {
            return inMeteor(function () {                              // 1418
              Grains.update(_this.grainId, { $set: { cachedViewInfo: viewInfo } });
            }).then(function () {                                      //
              return _this._callNewSession(request, viewInfo);         // 1421
            });                                                        //
          }, function (error) {                                        //
            if (error.kjType === "failed" || error.kjType === "unimplemented") {
              // Method not implemented.                               //
              // TODO(apibump): Don't treat 'failed' as 'unimplemented'. Unfortunately, old apps built
              //   with old versions of Cap'n Proto don't throw 'unimplemented' exceptions, so we have
              //   to accept 'failed' here at least until the next API bump.
              return _this._callNewSession(request, {});               // 1429
            } else {                                                   //
              return Promise.reject(error);                            // 1431
            }                                                          //
          });                                                          //
          _this5.session = new Capnp.Capability(promise, WebSession);  // 1434
        })();                                                          //
      }                                                                //
                                                                       //
      return this.session;                                             // 1437
    }                                                                  //
                                                                       //
    return getSession;                                                 //
  })();                                                                //
                                                                       //
  Proxy.prototype.keepAlive = (function () {                           // 1188
    function keepAlive() {                                             // 1440
      this.getConnection();                                            // 1441
      return this.supervisor.keepAlive();                              // 1442
    }                                                                  //
                                                                       //
    return keepAlive;                                                  //
  })();                                                                //
                                                                       //
  Proxy.prototype.resetConnection = (function () {                     // 1188
    function resetConnection() {                                       // 1445
      if (this.session) {                                              // 1446
        this.session.close();                                          // 1447
        delete this.session;                                           // 1448
      }                                                                //
                                                                       //
      if (this.uiView) {                                               // 1451
        this.uiView.close();                                           // 1452
        delete this.uiView;                                            // 1453
      }                                                                //
                                                                       //
      if (this.supervisor) {                                           // 1456
        this.supervisor.close();                                       // 1457
        delete this.supervisor;                                        // 1458
      }                                                                //
    }                                                                  //
                                                                       //
    return resetConnection;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.maybeRetryAfterError = (function () {                // 1188
    function maybeRetryAfterError(error, retryCount) {                 // 1462
      // If the error may be caused by the grain dying or a network failure, try to restart it,
      // returning a promise that resolves once restarted. Otherwise, just rethrow the error.
      // `retryCount` should be incremented for every successful retry as part of the same request;
      // we only want to retry once.                                   //
      var _this = this;                                                // 1467
      if (SandstormBackend.shouldRestartGrain(error, retryCount)) {    // 1468
        this.resetConnection();                                        // 1469
        return inMeteor(function () {                                  // 1470
          _this.supervisor = globalBackend.continueGrain(_this.grainId).supervisor;
        });                                                            //
      } else {                                                         //
        throw error;                                                   // 1474
      }                                                                //
    }                                                                  //
                                                                       //
    return maybeRetryAfterError;                                       //
  })();                                                                //
                                                                       //
  Proxy.prototype.doSessionInit = (function () {                       // 1188
    function doSessionInit(request, response, requestPath) {           // 1478
      // jscs:disable disallowQuotedKeysInObjects                      //
      var path = requestPath || "/";                                   // 1480
                                                                       //
      // Check that the path is relative (ie. starts with a /).        //
      // Also ensure that it doesn't start with 2 /, because that is interpreted as non-relative
      if (path.lastIndexOf("/", 0) !== 0 || path.lastIndexOf("//", 0) === 0) {
        response.writeHead(400, "Invalid path supplied", { "Content-Type": "text/plain" });
        response.end("Invalid path supplied.");                        // 1486
        return;                                                        // 1487
      }                                                                //
                                                                       //
      // Set the session ID.                                           //
      response.setHeader("Set-Cookie", ["sandstorm-sid=", this.sessionId, "; Max-Age=31536000; HttpOnly"].join(""));
                                                                       //
      response.setHeader("Cache-Control", "no-cache, private");        // 1493
                                                                       //
      // Redirect to the app's root URL.                               //
      // Note:  All browsers support relative locations and the next update to HTTP/1.1 will officially
      //   make them valid.  http://tools.ietf.org/html/draft-ietf-httpbis-p2-semantics-26#page-67
      response.writeHead(303, "See Other", { "Location": encodeURI(path) });
      response.end();                                                  // 1499
    }                                                                  //
                                                                       //
    return doSessionInit;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype.makeContext = (function () {                         // 1188
    function makeContext(request, response) {                          // 1502
      // Parses the cookies from the request, checks that the session ID is present and valid, then
      // returns the request context which contains the other cookies.  Throws an exception if the
      // session ID is missing or invalid.                             //
                                                                       //
      var context = {};                                                // 1507
                                                                       //
      if (this.hostId) {                                               // 1509
        var parseResult = parseCookies(request);                       // 1510
        if (!parseResult.sessionId || parseResult.sessionId !== this.sessionId) {
          throw new Meteor.Error(403, "Unauthorized");                 // 1512
        }                                                              //
                                                                       //
        if (parseResult.cookies.length > 0) {                          // 1515
          context.cookies = parseResult.cookies;                       // 1516
        }                                                              //
      } else {                                                         //
        // jscs:disable disallowEmptyBlocks                            //
        // This is an API request. Cookies are not supported.          //
      }                                                                //
                                                                       //
      context.accept = parseAcceptHeader(request);                     // 1523
                                                                       //
      context.eTagPrecondition = parsePreconditionHeader(request);     // 1525
                                                                       //
      context.additionalHeaders = [];                                  // 1527
      WebSession.Context.headerWhitelist.forEach(function (headerName) {
        if (request.headers[headerName]) {                             // 1529
          context.additionalHeaders.push({                             // 1530
            name: headerName,                                          // 1531
            value: request.headers[headerName]                         // 1532
          });                                                          //
        };                                                             //
      });                                                              //
                                                                       //
      var promise = new Promise(function (resolve, reject) {           // 1537
        response.resolveResponseStream = resolve;                      // 1538
        response.rejectResponseStream = reject;                        // 1539
      });                                                              //
                                                                       //
      context.responseStream = new Capnp.Capability(promise, ByteStream);
                                                                       //
      return context;                                                  // 1544
    }                                                                  //
                                                                       //
    return makeContext;                                                //
  })();                                                                //
                                                                       //
  Proxy.prototype.translateResponse = (function () {                   // 1188
    function translateResponse(rpcResponse, response, request) {       // 1547
      if (this.hostId) {                                               // 1548
        if (rpcResponse.setCookies && rpcResponse.setCookies.length > 0) {
          response.setHeader("Set-Cookie", rpcResponse.setCookies.map(makeSetCookieHeader));
        }                                                              //
                                                                       //
        // TODO(security): Add a Content-Security-Policy header which:
        // (1) Prevents the app from initiating HTTP requests to third parties.
        // (2) Prevents the app from navigating the parent frame.      //
        // (3) Prevents the app from opening popups.                   //
        // (4) Prohibits anyone other than the Sandstorm shell from framing the app (as a backup
        //   defense vs. clickjacking, though unguessable hostnames already mostly prevent this).
      } else {                                                         //
          // jscs:disable validateQuoteMarks                           //
          // This is an API request. Cookies are not supported.        //
                                                                       //
          // We need to make sure caches know that different bearer tokens get totally different results.
          response.setHeader('Vary', 'Authorization');                 // 1564
                                                                       //
          // APIs can be called from any origin. Because we ignore cookies, there is no security problem.
          response.setHeader('Access-Control-Allow-Origin', '*');      // 1567
                                                                       //
          // Add a Content-Security-Policy as a backup in case someone finds a way to load this resource
          // in a browser context. This policy should thoroughly neuter it.
          response.setHeader('Content-Security-Policy', "default-src 'none'; sandbox");
        }                                                              //
                                                                       //
      // On first response, update the session to have hasLoaded=true  //
      this.setHasLoaded();                                             // 1575
                                                                       //
      // TODO(security): Set X-Content-Type-Options: nosniff?          //
                                                                       //
      if ('content' in rpcResponse) {                                  // 1579
        var content = rpcResponse.content;                             // 1580
        var code = successCodes[content.statusCode];                   // 1581
        if (!code) {                                                   // 1582
          throw new Error('Unknown status code: ', content.statusCode);
        }                                                              //
                                                                       //
        if (content.mimeType) {                                        // 1586
          response.setHeader('Content-Type', content.mimeType);        // 1587
        }                                                              //
                                                                       //
        if (content.encoding) {                                        // 1590
          response.setHeader('Content-Encoding', content.encoding);    // 1591
        }                                                              //
                                                                       //
        if (content.language) {                                        // 1594
          response.setHeader('Content-Language', content.language);    // 1595
        }                                                              //
                                                                       //
        if (content.eTag) {                                            // 1598
          response.setHeader('ETag', composeETag(content.eTag));       // 1599
        }                                                              //
                                                                       //
        if ('disposition' in content && 'download' in content.disposition) {
          response.setHeader('Content-Disposition', 'attachment; filename="' + content.disposition.download.replace(/([\\"\n])/g, "\\$1") + '"');
        }                                                              //
                                                                       //
        if ('stream' in content.body) {                                // 1607
          if (request.method === 'HEAD') {                             // 1608
            content.body.stream.close();                               // 1609
            response.rejectResponseStream(new Error('HEAD request; content doesn\'t matter.'));
          } else {                                                     //
            var _ret3 = (function () {                                 //
              var streamHandle = content.body.stream;                  // 1612
              response.writeHead(code.id, code.title);                 // 1613
              var promise = new Promise(function (resolve, reject) {   // 1614
                response.resolveResponseStream(new Capnp.Capability(new ResponseStream(response, streamHandle, resolve, reject), ByteStream));
              });                                                      //
              promise.streamHandle = streamHandle;                     // 1618
              return {                                                 // 1619
                v: promise                                             //
              };                                                       //
            })();                                                      //
                                                                       //
            if (typeof _ret3 === "object") return _ret3.v;             //
          }                                                            //
        } else {                                                       //
          response.rejectResponseStream(new Error('Response content body was not a stream.'));
                                                                       //
          if ('bytes' in content.body) {                               // 1625
            response.setHeader('Content-Length', content.body.bytes.length);
          } else {                                                     //
            throw new Error('Unknown content body type.');             // 1628
          }                                                            //
        }                                                              //
                                                                       //
        response.writeHead(code.id, code.title);                       // 1632
                                                                       //
        if ('bytes' in content.body && request.method !== 'HEAD') {    // 1634
          response.write(content.body.bytes);                          // 1635
        }                                                              //
                                                                       //
        response.end();                                                // 1638
      } else if ('noContent' in rpcResponse) {                         //
        var noContent = rpcResponse.noContent;                         // 1640
        var noContentCode = noContentSuccessCodes[noContent.shouldResetForm * 1];
        response.writeHead(noContentCode.id, noContentCode.title);     // 1642
        response.end();                                                // 1643
      } else if ('preconditionFailed' in rpcResponse) {                //
        var preconditionFailed = rpcResponse.preconditionFailed;       // 1645
        if (request.method === 'GET' && 'if-none-match' in request.headers) {
          if (preconditionFailed.matchingETag) {                       // 1647
            response.setHeader('ETag', composeETag(preconditionFailed.matchingETag));
          }                                                            //
                                                                       //
          response.writeHead(304, 'Not Modified');                     // 1651
        } else {                                                       //
          response.writeHead(412, 'Precondition Failed');              // 1653
        }                                                              //
                                                                       //
        response.end();                                                // 1656
      } else if ('redirect' in rpcResponse) {                          //
        var redirect = rpcResponse.redirect;                           // 1658
        var redirectCode = redirectCodes[redirect.switchToGet * 2 + redirect.isPermanent];
        response.writeHead(redirectCode.id, redirectCode.title, {      // 1660
          'Location': redirect.location                                // 1661
        });                                                            //
        response.end();                                                // 1663
      } else if ('clientError' in rpcResponse) {                       //
        var clientError = rpcResponse.clientError;                     // 1665
        var errorCode = errorCodes[clientError.statusCode];            // 1666
        if (!errorCode) {                                              // 1667
          throw new Error('Unknown status code: ', clientError.statusCode);
        }                                                              //
                                                                       //
        response.writeHead(errorCode.id, errorCode.title, {            // 1671
          'Content-Type': 'text/html'                                  // 1672
        });                                                            //
                                                                       //
        if (request.method !== 'HEAD') {                               // 1675
          if (clientError.descriptionHtml) {                           // 1676
            response.write(clientError.descriptionHtml);               // 1677
          } else {                                                     //
            // TODO(someday):  Better default error page.              //
            response.write('<html><body><h1>' + errorCode.id + ': ' + errorCode.title + '</h1></body></html>');
          }                                                            //
        }                                                              //
                                                                       //
        response.end();                                                // 1685
      } else if ('serverError' in rpcResponse) {                       //
        response.writeHead(500, 'Internal Server Error', {             // 1687
          'Content-Type': 'text/html'                                  // 1688
        });                                                            //
                                                                       //
        if (request.method !== 'HEAD') {                               // 1691
          if (rpcResponse.serverError.descriptionHtml) {               // 1692
            response.write(rpcResponse.serverError.descriptionHtml);   // 1693
          } else {                                                     //
            // TODO(someday):  Better default error page.              //
            response.write('<html><body><h1>500: Internal Server Error</h1></body></html>');
          }                                                            //
        }                                                              //
                                                                       //
        response.end();                                                // 1700
      } else {                                                         //
        throw new Error('Unknown HTTP response type:\n' + JSON.stringify(rpcResponse));
      }                                                                //
                                                                       //
      return Promise.resolve(undefined);                               // 1705
    }                                                                  //
                                                                       //
    return translateResponse;                                          //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleRequest = (function () {                       // 1188
    function handleRequest(request, data, response, retryCount) {      // 1708
      var _this = this;                                                // 1709
      return Promise.resolve(undefined).then(function () {             // 1710
        return _this.makeContext(request, response);                   // 1711
      }).then(function (context) {                                     //
        // jscs:disable requireDotNotation                             //
        // Send the RPC.                                               //
        var path = request.url.slice(1); // remove leading '/'         // 1715
        var session = _this.getSession(request);                       // 1716
                                                                       //
        var requestContent = function () {                             // 1718
          return {                                                     // 1719
            content: data,                                             // 1720
            encoding: request.headers['content-encoding'],             // 1721
            mimeType: request.headers['content-type']                  // 1722
          };                                                           //
        };                                                             //
                                                                       //
        var xmlContent = function () {                                 // 1726
          var type = request.headers['content-type'] || 'application/xml;charset=utf-8';
          var match = type.match(/[^/]*\/xml(; *charset *= *([^ ;]*))?/);
          if (!match) {                                                // 1729
            response.writeHead(415, 'Unsupported media type.', {       // 1730
              'Content-Type': 'text/plain'                             // 1731
            });                                                        //
            response.end('expected XML request body');                 // 1733
            throw new Error('expected XML request body');              // 1734
          }                                                            //
                                                                       //
          var charset = match[2] || 'ISO-8859-1';                      // 1737
                                                                       //
          var encoding = request.headers['content-encoding'];          // 1739
          if (encoding && encoding !== 'identity') {                   // 1740
            if (encoding !== 'gzip') throw new Error('unknown Content-Encoding: ' + encoding);
            data = gunzipSync(data);                                   // 1742
          }                                                            //
                                                                       //
          return data.toString(charset.toLowerCase() === 'utf-8' ? 'utf8' : 'binary');
        };                                                             //
                                                                       //
        var propfindDepth = function () {                              // 1748
          var depth = request.headers['depth'];                        // 1749
          return depth === '0' ? 'zero' : depth === '1' ? 'one' : 'infinity';
        };                                                             //
                                                                       //
        var shallow = function () {                                    // 1755
          return request.headers['depth'] === '0';                     // 1756
        };                                                             //
                                                                       //
        var noOverwrite = function () {                                // 1759
          return (request.headers['overwrite'] || '').toLowerCase() === 'f';
        };                                                             //
                                                                       //
        var destination = function () {                                // 1763
          var result = request.headers['destination'];                 // 1764
          if (!result) throw new Error('missing destination');         // 1765
          return Url.parse(result).path.slice(1); // remove leading '/'
        };                                                             //
                                                                       //
        if (request.method === 'GET' || request.method === 'HEAD') {   // 1769
          return session.get(path, context, request.method === 'HEAD');
        } else if (request.method === 'POST') {                        //
          return session.post(path, requestContent(), context);        // 1772
        } else if (request.method === 'PUT') {                         //
          return session.put(path, requestContent(), context);         // 1774
        } else if (request.method === 'PATCH') {                       //
          return session.patch(path, requestContent(), context);       // 1776
        } else if (request.method === 'DELETE') {                      //
          return session["delete"](path, context);                     // 1778
        } else if (request.method === 'PROPFIND') {                    //
          return session.propfind(path, xmlContent(), propfindDepth(), context);
        } else if (request.method === 'PROPPATCH') {                   //
          return session.proppatch(path, xmlContent(), context);       // 1782
        } else if (request.method === 'MKCOL') {                       //
          return session.mkcol(path, requestContent(), context);       // 1784
        } else if (request.method === 'COPY') {                        //
          return session.copy(path, destination(), noOverwrite(), shallow(), context);
        } else if (request.method === 'MOVE') {                        //
          return session.move(path, destination(), noOverwrite(), context);
        } else if (request.method === 'LOCK') {                        //
          return session.lock(path, xmlContent(), shallow(), context);
        } else if (request.method === 'UNLOCK') {                      //
          return session.unlock(path, request.headers['lock-token'], context);
        } else if (request.method === 'ACL') {                         //
          return session.acl(path, xmlContent(), context);             // 1794
        } else if (request.method === 'REPORT') {                      //
          return session.report(path, requestContent(), context);      // 1796
        } else if (request.method === 'OPTIONS') {                     //
          return session.options(path, context).then(function (options) {
            var dav = [];                                              // 1799
            if (options.davClass1) dav.push('1');                      // 1800
            if (options.davClass2) dav.push('2');                      // 1801
            if (options.davClass3) dav.push('3');                      // 1802
            if (options.davExtensions) {                               // 1803
              options.davExtensions.forEach(function (token) {         // 1804
                if (token.match(/^([a-zA-Z0-9!#$%&'*+.^_`|~-]+|<[\x21-\x7E]*>)$/)) {
                  dav.push(token);                                     // 1806
                }                                                      //
              });                                                      //
            }                                                          //
                                                                       //
            if (dav.length > 0) {                                      // 1811
              response.setHeader("DAV", dav.join(", "));               // 1812
              response.setHeader("Access-Control-Expose-Headers", "DAV");
            }                                                          //
                                                                       //
            response.end();                                            // 1816
            // Return no response; we already handled everything.      //
          }, function (err) {                                          //
            if (err.kjType !== 'unimplemented') throw err;             // 1819
            response.end();                                            // 1820
            // Return no response; we already handled everything.      //
          });                                                          //
        } else {                                                       //
            throw new Error('Sandstorm only supports the following methods: GET, POST, PUT, PATCH, DELETE, HEAD, PROPFIND, PROPPATCH, MKCOL, COPY, MOVE, LOCK, UNLOCK, ACL, REPORT, and OPTIONS.');
          }                                                            //
      }).then(function (rpcResponse) {                                 //
        if (rpcResponse !== undefined) {                               // 1827
          // Will be undefined for OPTIONS request.                    //
          return _this.translateResponse(rpcResponse, response, request);
        }                                                              //
      })["catch"](function (error) {                                   //
        return _this.maybeRetryAfterError(error, retryCount).then(function () {
          return _this.handleRequest(request, data, response, retryCount + 1);
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleRequest;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleRequestStreaming = (function () {              // 1188
    function handleRequestStreaming(request, response, contentLength, retryCount) {
      var _this = this;                                                // 1838
      var context = this.makeContext(request, response);               // 1839
      var path = request.url.slice(1); // remove leading '/'           // 1840
      var session = this.getSession(request);                          // 1841
                                                                       //
      var mimeType = request.headers['content-type'] || 'application/octet-stream';
      var encoding = request.headers['content-encoding'];              // 1844
                                                                       //
      var requestStreamPromise = undefined;                            // 1846
      if (request.method === 'POST') {                                 // 1847
        requestStreamPromise = session.postStreaming(path, mimeType, context, encoding);
      } else if (request.method === 'PUT') {                           //
        requestStreamPromise = session.putStreaming(path, mimeType, context, encoding);
      } else {                                                         //
        throw new Error('Sandstorm only supports streaming POST and PUT requests.');
      }                                                                //
                                                                       //
      // TODO(perf): We ought to be pipelining the body, but we can't currently, because we have to
      //   handle the case where the app doesn't actually support streaming. We could pipeline while
      //   also buffering the data on the side in case we need it again later, but that's kind of
      //   complicated. We should fix the whole protocol to make streaming the standard.
      return requestStreamPromise.then(function (requestStreamResult) {
        var requestStream = requestStreamResult.stream;                // 1860
                                                                       //
        // Initialized when getResponse() returns, if the response is streaming.
        var downloadStreamHandle = undefined;                          // 1863
                                                                       //
        // Initialized if an upload-stream method throws.              //
        var uploadStreamError = undefined;                             // 1866
                                                                       //
        // We call `getResponse()` immediately so that the app can start streaming data down even while
        // data is still being streamed up. This theoretically allows apps to perform bidirectional
        // streaming, though probably very few actually do that.       //
        //                                                             //
        // Note that we need to be able to cancel `responsePromise` below, so it's important that it is
        // the raw Cap'n Proto promise. Hence `translateResponsePromise` is a separate variable.
        var responsePromise = requestStream.getResponse();             // 1874
                                                                       //
        var reportUploadStreamError = function (err) {                 // 1876
          // Called when an upload-stream method throws.               //
                                                                       //
          if (!uploadStreamError) {                                    // 1879
            uploadStreamError = err;                                   // 1880
                                                                       //
            // If we're still waiting on any response stuff, cancel it.
            responsePromise.cancel();                                  // 1883
            requestStream.close();                                     // 1884
            if (downloadStreamHandle) {                                // 1885
              downloadStreamHandle.close();                            // 1886
            }                                                          //
          }                                                            //
        };                                                             //
                                                                       //
        // If we have a Content-Length, pass it along to the app by calling `expectSize()`.
        if (contentLength !== undefined) {                             // 1892
          requestStream.expectSize(contentLength)["catch"](function (err) {
            // expectSize() is allowed to be unimplemented.            //
            if (err.kjType !== 'unimplemented') {                      // 1895
              reportUploadStreamError(err);                            // 1896
            }                                                          //
          });                                                          //
        }                                                              //
                                                                       //
        // Pipe the input stream to the app.                           //
        request.on('data', function (buf) {                            // 1902
          // TODO(soon): Only allow a small number of write()s to be in-flight at once,
          //   pausing the input stream if we hit that limit, so that we block the TCP socket all the
          //   way back to the source. May want to also coalesce small writes for this purpose.
          // TODO(security): The above problem may allow a DoS attack on the front-end.
          if (!uploadStreamError) requestStream.write(buf)["catch"](reportUploadStreamError);
        });                                                            //
                                                                       //
        request.on('end', function () {                                // 1910
          if (!uploadStreamError) requestStream.done()["catch"](reportUploadStreamError);
                                                                       //
          // We're all done making calls to requestStream.             //
          requestStream.close();                                       // 1914
        });                                                            //
                                                                       //
        request.on('close', function () {                              // 1917
          reportUploadStreamError(new Error('HTTP connection unexpectedly closed during request.'));
        });                                                            //
                                                                       //
        request.on('error', function (err) {                           // 1921
          reportUploadStreamError(err);                                // 1922
        });                                                            //
                                                                       //
        return responsePromise.then(function (rpcResponse) {           // 1925
          // Stop here if the upload stream has already failed.        //
          if (uploadStreamError) throw uploadStreamError;              // 1927
          var promise = _this.translateResponse(rpcResponse, response, request);
          downloadStreamHandle = promise.streamHandle;                 // 1929
          return promise;                                              // 1930
        });                                                            //
      }, function (err) {                                              //
        if (err.kjType === 'failed' && err.message.indexOf('not implemented') !== -1) {
          // Hack to work around old apps using an old version of Cap'n Proto, before the
          // 'unimplemented' exception type was introduced. :(         //
          // TODO(cleanup): When we transition to API version 2, we can move this into the
          //   compatibility layer.                                    //
          err.kjType = 'unimplemented';                                // 1938
        }                                                              //
                                                                       //
        if (SandstormBackend.shouldRestartGrain(err, 0)) {             // 1941
          // This is the kind of error that indicates we should retry. Note that we passed 0 for the
          // retry count above because we were just checking if this is a retriable error (vs. possibly
          // a method-not-implemented error); maybeRetryAfterError() will check again with the proper
          // retry count.                                              //
          return _this.maybeRetryAfterError(err, retryCount).then(function () {
            return _this.handleRequestStreaming(request, response, contentLength, retryCount + 1);
          });                                                          //
        } else if (err.kjType === 'unimplemented') {                   //
          // Streaming is not implemented. Fall back to non-streaming version.
          return readAll(request).then(function (data) {               // 1951
            return _this.handleRequest(request, data, response, 0);    // 1952
          });                                                          //
        } else {                                                       //
          throw err;                                                   // 1955
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleRequestStreaming;                                     //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleWebSocket = (function () {                     // 1188
    function handleWebSocket(request, socket, head, retryCount) {      // 1960
      var _this = this;                                                // 1961
                                                                       //
      return Promise.resolve(undefined).then(function () {             // 1963
        return _this.makeContext(request);                             // 1964
      }).then(function (context) {                                     //
        var path = request.url.slice(1); // remove leading '/'         // 1966
        var session = _this.getSession(request);                       // 1967
                                                                       //
        if (!('sec-websocket-key' in request.headers)) {               // 1969
          throw new Error('Missing Sec-WebSocket-Accept header.');     // 1970
        }                                                              //
                                                                       //
        var magic = request.headers['sec-websocket-key'] + '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
        var acceptKey = Crypto.createHash('sha1').update(magic).digest('base64');
                                                                       //
        var protocols = [];                                            // 1976
        if ('sec-websocket-protocol' in request.headers) {             // 1977
          protocols = request.headers['sec-websocket-protocol'].split(',').map(function (s) {
            return s.trim();                                           // 1979
          });                                                          //
        }                                                              //
                                                                       //
        var receiver = new WebSocketReceiver(socket);                  // 1982
                                                                       //
        var promise = session.openWebSocket(path, context, protocols, receiver);
                                                                       //
        if (head.length > 0) {                                         // 1986
          promise.serverStream.sendBytes(head);                        // 1987
        }                                                              //
                                                                       //
        var socketIdx = _this.websocketCounter.toString();             // 1990
        _this.websockets[socketIdx] = socket;                          // 1991
        _this.websocketCounter += 1;                                   // 1992
        pumpWebSocket(socket, promise.serverStream, function () {      // 1993
          delete _this.websockets[socketIdx];                          // 1993
        });                                                            //
                                                                       //
        return promise.then(function (response) {                      // 1995
          var headers = ['HTTP/1.1 101 Switching Protocols', 'Upgrade: websocket', 'Connection: Upgrade', 'Sec-WebSocket-Accept: ' + acceptKey];
          if (response.protocol && response.protocol.length > 0) {     // 2002
            headers.push('Sec-WebSocket-Protocol: ' + response.protocol.join(', '));
          }                                                            //
                                                                       //
          headers.push('');                                            // 2006
          headers.push('');                                            // 2007
                                                                       //
          socket.write(headers.join('\r\n'));                          // 2009
          receiver.go();                                               // 2010
                                                                       //
          // Note:  At this point errors are out of our hands.         //
        });                                                            //
      })["catch"](function (error) {                                   //
        return _this.maybeRetryAfterError(error, retryCount).then(function () {
          return _this.handleWebSocket(request, socket, head, retryCount + 1);
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleWebSocket;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.setHasLoaded = (function () {                        // 1188
    function setHasLoaded() {                                          // 2021
      var _this6 = this;                                               //
                                                                       //
      if (!this.hasLoaded) {                                           // 2022
        (function () {                                                 //
          _this6.hasLoaded = true;                                     // 2023
          var sessionId = _this6.sessionId;                            // 2024
          inMeteor(function () {                                       // 2025
            Sessions.update({ _id: sessionId }, { $set: { hasLoaded: true } });
          });                                                          //
        })();                                                          //
      }                                                                //
    }                                                                  //
                                                                       //
    return setHasLoaded;                                               //
  })();                                                                //
                                                                       //
  return Proxy;                                                        //
})();                                                                  //
                                                                       //
;                                                                      // 2030
                                                                       //
var PROTOCOL = Url.parse(process.env.ROOT_URL).protocol;               // 2032
                                                                       //
var isRfc1918OrLocal = function (address) {                            // 2034
  if (Net.isIPv4(address)) {                                           // 2035
    quad = address.split('.').map(function (x) {                       // 2036
      return parseInt(x, 10);                                          // 2036
    });                                                                //
    return quad[0] === 127 || quad[0] === 10 || quad[0] === 192 && quad[1] === 168 || quad[0] === 172 && quad[1] >= 16 && quad[1] < 32;
  } else if (Net.isIPv6(address)) {                                    //
    // IPv6 specifies ::1 as localhost and fd:: as reserved for private networks
    return address === '::1' || address.lastIndexOf('fd', 0) === 0;    // 2042
  } else {                                                             //
    // Ignore things that are neither IPv4 nor IPv6                    //
    return false;                                                      // 2045
  }                                                                    //
};                                                                     //
                                                                       //
var quadToIntString = function (quad) {                                // 2049
  var num = Bignum(quad[0]).shiftLeft(16).add(quad[1]).shiftLeft(16).add(quad[2]).shiftLeft(16).add(quad[3]);
  return num.toString();                                               // 2054
};                                                                     //
                                                                       //
// -----------------------------------------------------------------------------
// Session cookie management                                           //
                                                                       //
var parseCookies = function (request) {                                // 2060
  // jscs:disable requireDotNotation                                   //
  var header = request.headers['cookie'];                              // 2062
                                                                       //
  var result = { cookies: [] };                                        // 2064
  if (header) {                                                        // 2065
    var reqCookies = header.split(';');                                // 2066
    for (var i in babelHelpers.sanitizeForInObject(reqCookies)) {      // 2067
      var reqCookie = reqCookies[i];                                   // 2068
      var equalsPos = reqCookie.indexOf('=');                          // 2069
      var cookie = undefined;                                          // 2070
      if (equalsPos === -1) {                                          // 2071
        cookie = { key: reqCookie.trim(), value: '' };                 // 2072
      } else {                                                         //
        cookie = { key: reqCookie.slice(0, equalsPos).trim(), value: reqCookie.slice(equalsPos + 1) };
      }                                                                //
                                                                       //
      if (cookie.key === 'sandstorm-sid') {                            // 2077
        if (result.sessionId) {                                        // 2078
          throw new Error('Multiple sandstorm session IDs?');          // 2079
        }                                                              //
                                                                       //
        result.sessionId = cookie.value;                               // 2082
      } else {                                                         //
        result.cookies.push(cookie);                                   // 2084
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 2089
};                                                                     //
                                                                       //
var parsePreconditionHeader = function (request) {                     // 2092
  if (request.headers['if-match']) {                                   // 2093
    if (request.headers['if-match'].trim() === '*') {                  // 2094
      return { exists: null };                                         // 2095
    }                                                                  //
                                                                       //
    var matches = parseETagList(request.headers['if-match']);          // 2098
    if (matches.length > 0) {                                          // 2099
      return { matchesOneOf: matches };                                // 2100
    }                                                                  //
  }                                                                    //
                                                                       //
  if (request.headers['if-none-match']) {                              // 2104
    if (request.headers['if-none-match'].trim() === '*') {             // 2105
      return { doesntExist: null };                                    // 2106
    }                                                                  //
                                                                       //
    var noneMatches = parseETagList(request.headers['if-none-match']);
    if (noneMatches.length > 0) {                                      // 2110
      return { matchesNoneOf: noneMatches };                           // 2111
    }                                                                  //
  }                                                                    //
                                                                       //
  return { none: null };                                               // 2115
};                                                                     //
                                                                       //
var parseETagList = function (input) {                                 // 2118
  // An ETag is a quoted, \-escaped string, possibly prefixed with W/ (outside the quotes) to
  // indicate that it is weak. We are parsing a list of comma-delimited etags.
                                                                       //
  input = input.trim();                                                // 2122
  var results = [];                                                    // 2123
                                                                       //
  while (input.length > 0) {                                           // 2125
    var match = input.match(/^\s*(W\/)?"(([^"\\]|\\.)*)"\s*($|,)/);    // 2126
    if (!match) throw new Meteor.Error(400, 'invalid etag');           // 2127
                                                                       //
    input = input.slice(match[0].length).trim();                       // 2129
    results.push({ weak: !!match[1], value: match[2].replace(/\\(.)/g, '$1') });
  }                                                                    //
                                                                       //
  return results;                                                      // 2133
};                                                                     //
                                                                       //
var composeETag = function (tag) {                                     // 2136
  var result = '"' + (tag.value || '').replace(/([\\"])/g, '\\$1') + '"';
  if (tag.weak) result = 'W/' + result;                                // 2138
  return result;                                                       // 2139
};                                                                     //
                                                                       //
var parseAcceptHeader = function (request) {                           // 2142
  // jscs:disable requireDotNotation                                   //
  var header = request.headers['accept'];                              // 2144
                                                                       //
  var result = [];                                                     // 2146
  if (header) {                                                        // 2147
    var acceptList = header.split(',');                                // 2148
    for (var i in babelHelpers.sanitizeForInObject(acceptList)) {      // 2149
      var acceptStr = acceptList[i];                                   // 2150
      var tokensList = acceptStr.split(';');                           // 2151
                                                                       //
      var temp = { mimeType: tokensList[0].trim() };                   // 2153
                                                                       //
      var tokensListRest = tokensList.slice(1);                        // 2155
      for (var j in babelHelpers.sanitizeForInObject(tokensListRest)) {
        var token = tokensListRest[j];                                 // 2157
        var equalsPos = token.indexOf('=');                            // 2158
        if (equalsPos) {                                               // 2159
          var key = token.slice(0, equalsPos).trim();                  // 2160
          var value = token.slice(equalsPos + 1).trim();               // 2161
                                                                       //
          if (key === 'q') {                                           // 2163
            temp.qValue = +value;                                      // 2164
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      result.push(temp);                                               // 2169
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 2173
};                                                                     //
                                                                       //
// -----------------------------------------------------------------------------
// Regular HTTP request handling                                       //
                                                                       //
var readAll = function (stream) {                                      // 2179
  return new Promise(function (resolve, reject) {                      // 2180
    var buffers = [];                                                  // 2181
    stream.on('data', function (buf) {                                 // 2182
      buffers.push(buf);                                               // 2183
    });                                                                //
                                                                       //
    stream.on('end', function () {                                     // 2186
      resolve(Buffer.concat(buffers));                                 // 2187
    });                                                                //
                                                                       //
    stream.on('error', reject);                                        // 2190
  });                                                                  //
};                                                                     //
                                                                       //
var makeSetCookieHeader = function (cookie) {                          // 2194
  var result = [cookie.name, '=', cookie.value];                       // 2195
                                                                       //
  if ('absolute' in cookie.expires) {                                  // 2197
    result.push('; Expires=');                                         // 2198
    result.push(new Date(cookie.expires.absolute * 1000).toUTCString());
  } else if ('relative' in cookie.expires) {                           //
    result.push('; Max-Age=' + cookie.expires.relative);               // 2201
  }                                                                    //
                                                                       //
  if (cookie.path) {                                                   // 2204
    result.push('; Path=' + cookie.path);                              // 2205
  }                                                                    //
                                                                       //
  if (cookie.httpOnly) {                                               // 2208
    result.push('; HttpOnly');                                         // 2209
  }                                                                    //
                                                                       //
  return result.join('');                                              // 2212
};                                                                     //
                                                                       //
// TODO(cleanup):  Auto-generate based on annotations in web-session.capnp.
var successCodes = {                                                   // 2216
  ok: { id: 200, title: 'OK' },                                        // 2217
  created: { id: 201, title: 'Created' },                              // 2218
  accepted: { id: 202, title: 'Accepted' },                            // 2219
  multiStatus: { id: 207, title: 'Multi-Status' }                      // 2220
};                                                                     //
var noContentSuccessCodes = [                                          // 2222
// Indexed by shouldResetForm * 1                                      //
{ id: 204, title: 'No Content' }, { id: 205, title: 'Reset Content' }];
var redirectCodes = [                                                  // 2227
// Indexed by switchToGet * 2 + isPermanent                            //
{ id: 307, title: 'Temporary Redirect' }, { id: 308, title: 'Permanent Redirect' }, { id: 303, title: 'See Other' }, { id: 301, title: 'Moved Permanently' }];
var errorCodes = {                                                     // 2234
  badRequest: { id: 400, title: 'Bad Request' },                       // 2235
  forbidden: { id: 403, title: 'Forbidden' },                          // 2236
  notFound: { id: 404, title: 'Not Found' },                           // 2237
  methodNotAllowed: { id: 405, title: 'Method Not Allowed' },          // 2238
  notAcceptable: { id: 406, title: 'Not Acceptable' },                 // 2239
  conflict: { id: 409, title: 'Conflict' },                            // 2240
  gone: { id: 410, title: 'Gone' },                                    // 2241
  requestEntityTooLarge: { id: 413, title: 'Request Entity Too Large' },
  requestUriTooLong: { id: 414, title: 'Request-URI Too Long' },       // 2243
  unsupportedMediaType: { id: 415, title: 'Unsupported Media Type' },  // 2244
  imATeapot: { id: 418, title: 'I\'m a teapot' },                      // 2245
  unprocessableEntity: { id: 422, title: 'Unprocessable Entity' }      // 2246
};                                                                     //
                                                                       //
ResponseStream = (function () {                                        // 2249
  function ResponseStream(response, streamHandle, resolve, reject) {   // 2250
    babelHelpers.classCallCheck(this, ResponseStream);                 //
                                                                       //
    this.response = response;                                          // 2251
    this.streamHandle = streamHandle;                                  // 2252
    this.resolve = resolve;                                            // 2253
    this.reject = reject;                                              // 2254
    this.ended = false;                                                // 2255
  }                                                                    //
                                                                       //
  ResponseStream.prototype.write = (function () {                      // 2249
    function write(data) {                                             // 2258
      this.response.write(data);                                       // 2259
    }                                                                  //
                                                                       //
    return write;                                                      //
  })();                                                                //
                                                                       //
  ResponseStream.prototype.done = (function () {                       // 2249
    function done() {                                                  // 2262
      this.response.end();                                             // 2263
      this.streamHandle.close();                                       // 2264
      this.ended = true;                                               // 2265
    }                                                                  //
                                                                       //
    return done;                                                       //
  })();                                                                //
                                                                       //
  ResponseStream.prototype.close = (function () {                      // 2249
    function close() {                                                 // 2268
      if (this.ended) {                                                // 2269
        this.resolve();                                                // 2270
      } else {                                                         //
        this.streamHandle.close();                                     // 2272
        this.reject(new Error('done() was never called on outbound stream.'));
      }                                                                //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  return ResponseStream;                                               //
})();                                                                  //
                                                                       //
// TODO(cleanup): Node 0.12 has a `gunzipSync` but 0.10 (which Meteor still uses) does not.
var Zlib = Npm.require('zlib');                                        // 2279
var gunzipSync = Meteor.wrapAsync(Zlib.gunzip, Zlib);                  // 2280
                                                                       //
// -----------------------------------------------------------------------------
// WebSocket handling                                                  //
                                                                       //
WebSocketReceiver = (function () {                                     // 2285
  function WebSocketReceiver(socket) {                                 // 2286
    babelHelpers.classCallCheck(this, WebSocketReceiver);              //
                                                                       //
    this.socket = socket;                                              // 2287
    this.queue = [];                                                   // 2288
  }                                                                    //
                                                                       //
  WebSocketReceiver.prototype.go = (function () {                      // 2285
    function go() {                                                    // 2291
      for (var i in babelHelpers.sanitizeForInObject(this.queue)) {    // 2292
        this.socket.write(this.queue[i]);                              // 2293
      }                                                                //
                                                                       //
      this.queue = null;                                               // 2296
    }                                                                  //
                                                                       //
    return go;                                                         //
  })();                                                                //
                                                                       //
  WebSocketReceiver.prototype.sendBytes = (function () {               // 2285
    function sendBytes(message) {                                      // 2299
      // TODO(someday):  Flow control of some sort?                    //
      if (this.queue === null) {                                       // 2301
        this.socket.write(message);                                    // 2302
      } else {                                                         //
        this.queue.push(message);                                      // 2304
      }                                                                //
    }                                                                  //
                                                                       //
    return sendBytes;                                                  //
  })();                                                                //
                                                                       //
  WebSocketReceiver.prototype.close = (function () {                   // 2285
    function close() {                                                 // 2308
      this.socket.end();                                               // 2309
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  return WebSocketReceiver;                                            //
})();                                                                  //
                                                                       //
pumpWebSocket = function (socket, rpcStream, destructor) {             // 2313
  socket.on('data', function (chunk) {                                 // 2314
    rpcStream.sendBytes(chunk)["catch"](function (err) {               // 2315
      if (err.kjType !== 'disconnected') {                             // 2316
        console.error('WebSocket sendBytes failed: ' + err.stack);     // 2317
      }                                                                //
                                                                       //
      socket.destroy();                                                // 2320
    });                                                                //
  });                                                                  //
                                                                       //
  socket.on('end', function (chunk) {                                  // 2324
    rpcStream.close();                                                 // 2325
  });                                                                  //
                                                                       //
  socket.on("close", function () {                                     // 2328
    destructor();                                                      // 2329
  });                                                                  //
};                                                                     //
                                                                       //
// =======================================================================================
// Debug log access                                                    //
                                                                       //
Meteor.publish('grainLog', function (grainId) {                        // 2336
  var _this7 = this;                                                   //
                                                                       //
  check(grainId, String);                                              // 2337
  var id = 0;                                                          // 2338
  var grain = Grains.findOne(grainId);                                 // 2339
  if (!grain || !this.userId || grain.userId !== this.userId) {        // 2340
    this.added('grainLog', id++, { text: 'Only the grain owner can view the debug log.' });
    this.ready();                                                      // 2342
    return;                                                            // 2343
  }                                                                    //
                                                                       //
  var connected = false;                                               // 2346
  var _this = this;                                                    // 2347
                                                                       //
  var receiver = {                                                     // 2349
    write: function (data) {                                           // 2350
      connected = true;                                                // 2351
      _this.added('grainLog', id++, { text: data.toString('utf8') });  // 2352
    },                                                                 //
                                                                       //
    close: function () {                                               // 2355
      if (connected) {                                                 // 2356
        _this.added('grainLog', id++, {                                // 2357
          text: '*** lost connection to grain (probably because it shut down) ***'
        });                                                            //
      }                                                                //
    }                                                                  //
  };                                                                   //
                                                                       //
  try {                                                                // 2364
    (function () {                                                     //
      var handle = waitPromise(globalBackend.useGrain(grainId, function (supervisor) {
        return supervisor.watchLog(8192, receiver);                    // 2366
      })).handle;                                                      //
      connected = true;                                                // 2368
      _this7.onStop(function () {                                      // 2369
        handle.close();                                                // 2370
      });                                                              //
    })();                                                              //
  } catch (err) {                                                      //
    if (!connected) {                                                  // 2373
      this.added('grainLog', id++, {                                   // 2374
        text: '*** couldn\'t connect to grain (' + err + ') ***'       // 2375
      });                                                              //
    }                                                                  //
  }                                                                    //
                                                                       //
  // Notify ready.                                                     //
  this.ready();                                                        // 2381
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=proxy.js.map
