(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var RoutePolicy = Package.routepolicy.RoutePolicy;
var WebApp = Package.webapp.WebApp;
var main = Package.webapp.main;
var WebAppInternals = Package.webapp.WebAppInternals;
var _ = Package.underscore._;
var ServiceConfiguration = Package['service-configuration'].ServiceConfiguration;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var Random = Package.random.Random;
var ECMAScript = Package.ecmascript.ECMAScript;
var babelHelpers = Package['babel-runtime'].babelHelpers;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var Promise = Package.promise.Promise;

/* Package-scope variables */
var middleware, _saml, SAML, profile;

(function(){

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                  //
// packages/accounts-saml/saml_server.js                                                                            //
//                                                                                                                  //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                    //
if (!Accounts.saml) {                                                                                               // 1
  Accounts.saml = {};                                                                                               // 2
}                                                                                                                   //
                                                                                                                    //
var Url = Npm.require("url");                                                                                       // 5
var Fiber = Npm.require("fibers");                                                                                  // 6
var connect = Npm.require("connect");                                                                               // 7
                                                                                                                    //
RoutePolicy.declare("/_saml/", "network");                                                                          // 9
                                                                                                                    //
var HOSTNAME = Url.parse(process.env.ROOT_URL).hostname;                                                            // 11
                                                                                                                    //
Accounts.registerLoginHandler(function (loginRequest) {                                                             // 13
  if (!loginRequest.saml || !loginRequest.credentialToken) {                                                        // 14
    return undefined;                                                                                               // 15
  }                                                                                                                 //
                                                                                                                    //
  if (!Accounts.identityServices.saml.isEnabled()) {                                                                // 18
    throw new Meteor.Error(403, "SAML service is disabled.");                                                       // 19
  }                                                                                                                 //
                                                                                                                    //
  var loginResult = Accounts.saml.retrieveCredential(loginRequest.credentialToken);                                 // 22
  if (!loginResult) {                                                                                               // 23
    throw new Meteor.Error(500, "SAML login did not complete.");                                                    // 24
  } else if (loginResult.profile && loginResult.profile.email) {                                                    //
    var user = _.pick(loginResult.profile, "displayName", "email");                                                 // 26
    user.id = loginResult.profile.nameID;                                                                           // 27
    return Accounts.updateOrCreateUserFromExternalService("saml", user, {});                                        // 28
  } else {                                                                                                          //
    throw new Meteor.Error(500, "SAML profile did not contain an email address");                                   // 30
  }                                                                                                                 //
});                                                                                                                 //
                                                                                                                    //
// TODO(soon): This may need to be a Mongo collection in order to work when the frontend is                         //
//   replicated (but currently Sandstorm for Work is not replicated).                                               //
Accounts.saml._loginResultForCredentialToken = {};                                                                  // 36
                                                                                                                    //
Accounts.saml.hasCredential = function (credentialToken) {                                                          // 38
  return _.has(Accounts.saml._loginResultForCredentialToken, credentialToken);                                      // 39
};                                                                                                                  //
                                                                                                                    //
Accounts.saml.retrieveCredential = function (credentialToken) {                                                     // 42
  var result = Accounts.saml._loginResultForCredentialToken[credentialToken];                                       // 43
  delete Accounts.saml._loginResultForCredentialToken[credentialToken];                                             // 44
  return result;                                                                                                    // 45
};                                                                                                                  //
                                                                                                                    //
// Listen to incoming OAuth http requests                                                                           //
WebApp.connectHandlers.use(connect.bodyParser()).use(function (req, res, next) {                                    // 49
  // Need to create a Fiber since we're using synchronous http calls and nothing                                    //
  // else is wrapping this in a fiber automatically                                                                 //
  Fiber(function () {                                                                                               // 52
    middleware(req, res, next);                                                                                     // 53
  }).run();                                                                                                         //
});                                                                                                                 //
                                                                                                                    //
middleware = function (req, res, next) {                                                                            // 57
  // Make sure to catch any exceptions because otherwise we'd crash                                                 //
  // the runner                                                                                                     //
  try {                                                                                                             // 60
    if (!Accounts.identityServices.saml.isEnabled()) {                                                              // 61
      next();                                                                                                       // 62
      return;                                                                                                       // 63
    }                                                                                                               //
                                                                                                                    //
    var samlObject = samlUrlToObject(req.url);                                                                      // 66
    if (!samlObject || !samlObject.serviceName) {                                                                   // 67
      next();                                                                                                       // 68
      return;                                                                                                       // 69
    }                                                                                                               //
                                                                                                                    //
    if (!samlObject.actionName) throw new Error("Missing SAML action");                                             // 72
                                                                                                                    //
    var service = {                                                                                                 // 75
      "provider": "default",                                                                                        // 76
      "entryPoint": SandstormDb.prototype.getSamlEntryPoint(),                                                      // 77
      // TODO(someday): find a better way to inject the DB                                                          //
      "issuer": HOSTNAME,                                                                                           // 79
      "cert": SandstormDb.prototype.getSamlPublicCert()                                                             // 80
    };                                                                                                              //
                                                                                                                    //
    // Skip everything if there's no service set by the saml middleware                                             //
    if (!service || samlObject.serviceName !== service.provider) throw new Error("Unexpected SAML service " + samlObject.serviceName);
                                                                                                                    //
    if (samlObject.actionName === "authorize") {                                                                    // 87
      service.callbackUrl = Meteor.absoluteUrl("_saml/validate/" + service.provider);                               // 88
      service.id = samlObject.credentialToken;                                                                      // 89
      _saml = new SAML(service);                                                                                    // 90
      _saml.getAuthorizeUrl(req, function (err, url) {                                                              // 91
        if (err) throw new Error("Unable to generate authorize url");                                               // 92
        res.writeHead(302, { "Location": url });                                                                    // 94
        res.end();                                                                                                  // 95
      });                                                                                                           //
    } else if (samlObject.actionName === "validate") {                                                              //
      _saml = new SAML(service);                                                                                    // 98
      _saml.validateResponse(req.body.SAMLResponse, function (err, profile, loggedOut) {                            // 99
        if (err) {                                                                                                  // 100
          console.error("Error validating SAML response", err.toString());                                          // 101
          throw new Error("Unable to validate response url");                                                       // 102
        }                                                                                                           //
                                                                                                                    //
        // Do NOT use samlObject.credentialToken; it isn't signed!                                                  //
        var credentialToken = profile.inResponseToId || profile.InResponseTo;                                       // 106
        if (!credentialToken) {                                                                                     // 107
          throw new Error("SAML response missing InResponseTo attribute. Sandstorm does not support " + "IdP-initiated authentication; authentication requests must start " + "from the user choosing SAML login in the Sandstorm UI.");
        }                                                                                                           //
                                                                                                                    //
        Accounts.saml._loginResultForCredentialToken[credentialToken] = {                                           // 114
          profile: profile                                                                                          // 115
        };                                                                                                          //
                                                                                                                    //
        closePopup(res);                                                                                            // 118
      });                                                                                                           //
    } else {                                                                                                        //
      throw new Error("Unexpected SAML action " + samlObject.actionName);                                           // 121
    }                                                                                                               //
  } catch (err) {                                                                                                   //
    closePopup(res, err);                                                                                           // 124
  }                                                                                                                 //
};                                                                                                                  //
                                                                                                                    //
var samlUrlToObject = function (url) {                                                                              // 128
  // req.url will be "/_saml/<action>/<service name>/<credentialToken>"                                             //
  if (!url) return null;                                                                                            // 130
                                                                                                                    //
  var splitPath = url.split("/");                                                                                   // 133
                                                                                                                    //
  // Any non-saml request will continue down the default                                                            //
  // middlewares.                                                                                                   //
  if (splitPath[1] !== "_saml") return null;                                                                        // 137
                                                                                                                    //
  return {                                                                                                          // 140
    actionName: splitPath[2],                                                                                       // 141
    serviceName: splitPath[3],                                                                                      // 142
    credentialToken: splitPath[4]                                                                                   // 143
  };                                                                                                                //
};                                                                                                                  //
                                                                                                                    //
var closePopup = function (res, err) {                                                                              // 147
  res.writeHead(200, { "Content-Type": "text/html" });                                                              // 148
  var content = "<html><head><script>window.close()</script></head></html>";                                        // 149
  if (err) content = "<html><body><h2>Sorry, an error occured</h2><div>" + err + '</div><a onclick="window.close();">Close Window</a></body></html>';
  res.end(content, "utf-8");                                                                                        // 153
};                                                                                                                  //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                  //
// packages/accounts-saml/saml_utils.js                                                                             //
//                                                                                                                  //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                    //
var zlib = Npm.require("zlib");                                                                                     // 1
var xml2js = Npm.require("xml2js");                                                                                 // 2
var xmlCrypto = Npm.require("xml-crypto");                                                                          // 3
var crypto = Npm.require("crypto");                                                                                 // 4
var xmldom = Npm.require("xmldom");                                                                                 // 5
var querystring = Npm.require("querystring");                                                                       // 6
                                                                                                                    //
SAML = function (options) {                                                                                         // 8
  this.options = this.initialize(options);                                                                          // 9
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.initialize = function (options) {                                                                    // 12
  if (!options) {                                                                                                   // 13
    options = {};                                                                                                   // 14
  }                                                                                                                 //
                                                                                                                    //
  if (!options.protocol) {                                                                                          // 17
    options.protocol = "https://";                                                                                  // 18
  }                                                                                                                 //
                                                                                                                    //
  if (!options.path) {                                                                                              // 21
    options.path = "/saml/consume";                                                                                 // 22
  }                                                                                                                 //
                                                                                                                    //
  if (!options.issuer) {                                                                                            // 25
    options.issuer = "onelogin_saml";                                                                               // 26
  }                                                                                                                 //
                                                                                                                    //
  if (options.identifierFormat === undefined) {                                                                     // 29
    options.identifierFormat = "urn:oasis:names:tc:SAML:2.0:nameid-format:persistent";                              // 30
  }                                                                                                                 //
                                                                                                                    //
  return options;                                                                                                   // 33
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.generateUniqueID = function () {                                                                     // 36
  var chars = "abcdef0123456789";                                                                                   // 37
  var uniqueID = "";                                                                                                // 38
  for (var i = 0; i < 20; i++) {                                                                                    // 39
    uniqueID += chars.substr(Math.floor(Math.random() * 15), 1);                                                    // 40
  }                                                                                                                 //
                                                                                                                    //
  return uniqueID;                                                                                                  // 43
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.generateInstant = function () {                                                                      // 46
  var date = new Date();                                                                                            // 47
  return date.getUTCFullYear() + "-" + ("0" + (date.getUTCMonth() + 1)).slice(-2) + "-" + ("0" + date.getUTCDate()).slice(-2) + "T" + ("0" + date.getUTCHours()).slice(-2) + ":" + ("0" + date.getUTCMinutes()).slice(-2) + ":" + ("0" + date.getUTCSeconds()).slice(-2) + "Z";
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.signRequest = function (xml) {                                                                       // 51
  var signer = crypto.createSign("RSA-SHA1");                                                                       // 52
  signer.update(xml);                                                                                               // 53
  return signer.sign(this.options.privateCert, "base64");                                                           // 54
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.generateAuthorizeRequest = function (req) {                                                          // 57
  var id = "_" + this.generateUniqueID();                                                                           // 58
  var instant = this.generateInstant();                                                                             // 59
                                                                                                                    //
  var callbackUrl = undefined;                                                                                      // 61
  // Post-auth destination                                                                                          //
  if (this.options.callbackUrl) {                                                                                   // 63
    callbackUrl = this.options.callbackUrl;                                                                         // 64
  } else {                                                                                                          //
    callbackUrl = this.options.protocol + req.headers.host + this.options.path;                                     // 66
  }                                                                                                                 //
                                                                                                                    //
  if (this.options.id) id = this.options.id;                                                                        // 69
                                                                                                                    //
  var request = "<samlp:AuthnRequest xmlns:samlp=\"urn:oasis:names:tc:SAML:2.0:protocol\" ID=\"" + id + "\" Version=\"2.0\" IssueInstant=\"" + instant + "\" ProtocolBinding=\"urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST\" AssertionConsumerServiceURL=\"" + callbackUrl + "\" Destination=\"" + this.options.entryPoint + "\">" + "<saml:Issuer xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\">" + this.options.issuer + "</saml:Issuer>\n";
                                                                                                                    //
  if (this.options.identifierFormat) {                                                                              // 78
    request += "<samlp:NameIDPolicy xmlns:samlp=\"urn:oasis:names:tc:SAML:2.0:protocol\" Format=\"" + this.options.identifierFormat + "\" AllowCreate=\"true\"></samlp:NameIDPolicy>\n";
  }                                                                                                                 //
                                                                                                                    //
  request += "<samlp:RequestedAuthnContext xmlns:samlp=\"urn:oasis:names:tc:SAML:2.0:protocol\" Comparison=\"exact\">" + "<saml:AuthnContextClassRef xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\">urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport</saml:AuthnContextClassRef></samlp:RequestedAuthnContext>\n" + "</samlp:AuthnRequest>";
                                                                                                                    //
  return request;                                                                                                   // 88
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.generateLogoutRequest = function (req) {                                                             // 91
  var id = "_" + this.generateUniqueID();                                                                           // 92
  var instant = this.generateInstant();                                                                             // 93
                                                                                                                    //
  //samlp:LogoutRequest xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
  // ID="_135ad2fd-b275-4428-b5d6-3ac3361c3a7f" Version="2.0" Destination="https://idphost/adfs/ls/"                //
  //IssueInstant="2008-06-03T12:59:57Z"><saml:Issuer>myhost</saml:Issuer><NameID Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
  //NameQualifier="https://idphost/adfs/ls/">myemail@mydomain.com</NameID<samlp:SessionIndex>_0628125f-7f95-42cc-ad8e-fde86ae90bbe
  //</samlp:SessionIndex></samlp:LogoutRequest>                                                                     //
                                                                                                                    //
  var request = "<samlp:LogoutRequest xmlns:samlp=\"urn:oasis:names:tc:SAML:2.0:protocol\" " + "xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\" ID=\"" + id + "\" Version=\"2.0\" IssueInstant=\"" + instant + "\" Destination=\"" + this.options.entryPoint + "\">" + "<saml:Issuer xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\">" + this.options.issuer + "</saml:Issuer>" + "<saml:NameID Format=\"" + req.user.nameIDFormat + "\">" + req.user.nameID + "</saml:NameID>" + "</samlp:LogoutRequest>";
  return request;                                                                                                   // 107
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.requestToUrl = function (request, operation, callback) {                                             // 110
  var _this = this;                                                                                                 // 111
  zlib.deflateRaw(request, function (err, buffer) {                                                                 // 112
    if (err) {                                                                                                      // 113
      return callback(err);                                                                                         // 114
    }                                                                                                               //
                                                                                                                    //
    var base64 = buffer.toString("base64");                                                                         // 117
    var target = _this.options.entryPoint;                                                                          // 118
                                                                                                                    //
    if (operation === "logout") {                                                                                   // 120
      if (_this.options.logoutUrl) {                                                                                // 121
        target = _this.options.logoutUrl;                                                                           // 122
      }                                                                                                             //
    }                                                                                                               //
                                                                                                                    //
    if (target.indexOf("?") > 0) target += "&";else target += "?";                                                  // 126
                                                                                                                    //
    var samlRequest = {                                                                                             // 131
      SAMLRequest: base64                                                                                           // 132
    };                                                                                                              //
                                                                                                                    //
    if (_this.options.privateCert) {                                                                                // 135
      samlRequest.SigAlg = "http://www.w3.org/2000/09/xmldsig#rsa-sha1";                                            // 136
      samlRequest.Signature = _this.signRequest(querystring.stringify(samlRequest));                                // 137
    }                                                                                                               //
                                                                                                                    //
    target += querystring.stringify(samlRequest);                                                                   // 140
                                                                                                                    //
    callback(null, target);                                                                                         // 142
  });                                                                                                               //
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.getAuthorizeUrl = function (req, callback) {                                                         // 146
  var request = this.generateAuthorizeRequest(req);                                                                 // 147
                                                                                                                    //
  this.requestToUrl(request, "authorize", callback);                                                                // 149
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.getLogoutUrl = function (req, callback) {                                                            // 152
  var request = this.generateLogoutRequest(req);                                                                    // 153
                                                                                                                    //
  this.requestToUrl(request, "logout", callback);                                                                   // 155
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.certToPEM = function (cert) {                                                                        // 158
  cert = cert.match(/.{1,64}/g).join("\n");                                                                         // 159
  cert = "-----BEGIN CERTIFICATE-----\n" + cert;                                                                    // 160
  cert = cert + "\n-----END CERTIFICATE-----\n";                                                                    // 161
  return cert;                                                                                                      // 162
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.validateSignature = function (xml, cert) {                                                           // 165
  var _this = this;                                                                                                 // 166
  var doc = new xmldom.DOMParser().parseFromString(xml);                                                            // 167
  var signature = xmlCrypto.xpath(doc, "//*[local-name(.)='Signature' and namespace-uri(.)='http://www.w3.org/2000/09/xmldsig#']")[0];
  var sig = new xmlCrypto.SignedXml();                                                                              // 169
  sig.keyInfoProvider = {                                                                                           // 170
    getKeyInfo: function (key) {                                                                                    // 171
      return "<X509Data></X509Data>";                                                                               // 172
    },                                                                                                              //
                                                                                                                    //
    getKey: function (keyInfo) {                                                                                    // 175
      return _this.certToPEM(cert);                                                                                 // 176
    }                                                                                                               //
  };                                                                                                                //
  sig.loadSignature(signature.toString());                                                                          // 179
  return sig.checkSignature(xml);                                                                                   // 180
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.getElement = function (parentElement, elementName) {                                                 // 183
  if (parentElement["saml:" + elementName]) {                                                                       // 184
    return parentElement["saml:" + elementName];                                                                    // 185
  } else if (parentElement["samlp:" + elementName]) {                                                               //
    return parentElement["samlp:" + elementName];                                                                   // 187
  } else if (parentElement["saml2p:" + elementName]) {                                                              //
    return parentElement["saml2p:" + elementName];                                                                  // 189
  } else if (parentElement["saml2:" + elementName]) {                                                               //
    return parentElement["saml2:" + elementName];                                                                   // 191
  }                                                                                                                 //
                                                                                                                    //
  return parentElement[elementName];                                                                                // 194
};                                                                                                                  //
                                                                                                                    //
SAML.prototype.validateResponse = function (samlResponse, callback) {                                               // 197
  var _this = this;                                                                                                 // 198
  var xml = new Buffer(samlResponse, "base64").toString("ascii");                                                   // 199
  var parser = new xml2js.Parser({ explicitRoot: true });                                                           // 200
  parser.parseString(xml, function (err, doc) {                                                                     // 201
    // Verify signature                                                                                             //
    if (_this.options.cert && !_this.validateSignature(xml, _this.options.cert)) {                                  // 203
      return callback(new Error("Invalid signature"), null, false);                                                 // 204
    }                                                                                                               //
                                                                                                                    //
    var response = _this.getElement(doc, "Response");                                                               // 207
    if (response) {                                                                                                 // 208
      var assertion = _this.getElement(response, "Assertion");                                                      // 209
      if (!assertion) {                                                                                             // 210
        return callback(new Error("Missing SAML assertion"), null, false);                                          // 211
      }                                                                                                             //
                                                                                                                    //
      profile = {};                                                                                                 // 214
                                                                                                                    //
      if (response.$ && response.$.InResponseTo) {                                                                  // 216
        profile.inResponseToId = response.$.InResponseTo;                                                           // 217
      }                                                                                                             //
                                                                                                                    //
      if (response.$ && response.$.Destination) {                                                                   // 220
        if (!response.$.Destination.startsWith(process.env.ROOT_URL)) {                                             // 221
          return callback(new Error("SAML Response received with invalid Destination: " + response.$.Destination));
        }                                                                                                           //
      }                                                                                                             //
                                                                                                                    //
      var issuer = _this.getElement(assertion[0], "Issuer");                                                        // 227
      if (issuer) {                                                                                                 // 228
        profile.issuer = issuer[0];                                                                                 // 229
      }                                                                                                             //
                                                                                                                    //
      var subject = _this.getElement(assertion[0], "Subject");                                                      // 232
      if (subject) {                                                                                                // 233
        var nameID = _this.getElement(subject[0], "NameID");                                                        // 234
        if (nameID) {                                                                                               // 235
          profile.nameID = nameID[0]._;                                                                             // 236
                                                                                                                    //
          if (nameID[0].$.Format) {                                                                                 // 238
            profile.nameIDFormat = nameID[0].$.Format;                                                              // 239
            if (profile.nameIDFormat.toLowerCase().indexOf("transient") !== -1) {                                   // 240
              return callback(new Error("SAML returned a transient NameID. Sandstorm requires a persistent NameID. " + "Please check your IdP config."));
            }                                                                                                       //
          }                                                                                                         //
        }                                                                                                           //
                                                                                                                    //
        var subjectConfirmation = _this.getElement(subject[0], "SubjectConfirmation");                              // 248
        if (subjectConfirmation) {                                                                                  // 249
          var subjectConfirmationData = _this.getElement(subjectConfirmation[0], "SubjectConfirmationData")[0];     // 250
          if (subjectConfirmationData) {                                                                            // 252
            var recipient = subjectConfirmationData.$.Recipient;                                                    // 253
            if (recipient && !recipient.startsWith(process.env.ROOT_URL)) {                                         // 254
              return callback(new Error("SAML sent to wrong recipient"));                                           // 255
            }                                                                                                       //
                                                                                                                    //
            var nowMs = Date.now();                                                                                 // 258
            var notOnOrBefore = subjectConfirmationData.$.NotOnOrBefore;                                            // 259
            if (notOnOrBefore && nowMs <= Date.parse(notOnOrBefore)) {                                              // 260
              return callback(new Error("SAML assertion was signed for the future."));                              // 261
            }                                                                                                       //
                                                                                                                    //
            var notOnOrAfter = subjectConfirmationData.$.NotOnOrAfter;                                              // 264
            if (notOnOrAfter && nowMs >= Date.parse(notOnOrAfter)) {                                                // 265
              return callback(new Error("SAML assertion was signed for the past."));                                // 266
            }                                                                                                       //
          }                                                                                                         //
        }                                                                                                           //
      }                                                                                                             //
                                                                                                                    //
      var attributeStatement = _this.getElement(assertion[0], "AttributeStatement");                                // 272
      if (attributeStatement) {                                                                                     // 273
        var attributes = _this.getElement(attributeStatement[0], "Attribute");                                      // 274
                                                                                                                    //
        if (attributes) {                                                                                           // 276
          attributes.forEach(function (attribute) {                                                                 // 277
            var value = _this.getElement(attribute, "AttributeValue");                                              // 278
            if (typeof value[0] === "string") {                                                                     // 279
              profile[attribute.$.Name] = value[0];                                                                 // 280
            } else {                                                                                                //
              profile[attribute.$.Name] = value[0]._;                                                               // 282
            }                                                                                                       //
          });                                                                                                       //
        }                                                                                                           //
                                                                                                                    //
        if (!profile.mail && profile["urn:oid:0.9.2342.19200300.100.1.3"]) {                                        // 287
          // See http://www.incommonfederation.org/attributesummary.html for definition of attribute OIDs           //
          profile.mail = profile["urn:oid:0.9.2342.19200300.100.1.3"];                                              // 289
        }                                                                                                           //
                                                                                                                    //
        if (!profile.email && profile.mail) {                                                                       // 292
          profile.email = profile.mail;                                                                             // 293
        }                                                                                                           //
      }                                                                                                             //
                                                                                                                    //
      if (!profile.email && profile.nameID && profile.nameIDFormat && profile.nameIDFormat.indexOf("emailAddress") >= 0) {
        profile.email = profile.nameID;                                                                             // 298
      }                                                                                                             //
                                                                                                                    //
      callback(null, profile, false);                                                                               // 301
    } else {                                                                                                        //
      var logoutResponse = _this.getElement(doc, "LogoutResponse");                                                 // 303
                                                                                                                    //
      if (logoutResponse) {                                                                                         // 305
        callback(null, null, true);                                                                                 // 306
      } else {                                                                                                      //
        return callback(new Error("Unknown SAML response message"), null, false);                                   // 308
      }                                                                                                             //
    }                                                                                                               //
  });                                                                                                               //
};                                                                                                                  //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['accounts-saml'] = {};

})();

//# sourceMappingURL=accounts-saml.js.map
