(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var ECMAScript = Package.ecmascript.ECMAScript;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var babelHelpers = Package['babel-runtime'].babelHelpers;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var Promise = Package.promise.Promise;

/* Package-scope variables */
var ContactProfiles;

(function(){

/////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                         //
// packages/sandstorm-contacts/contacts-server.js                                          //
//                                                                                         //
/////////////////////////////////////////////////////////////////////////////////////////////
                                                                                           //
// Sandstorm - Personal Cloud Sandbox                                                      //
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                   //
// All rights reserved.                                                                    //
//                                                                                         //
// Licensed under the Apache License, Version 2.0 (the "License");                         //
// you may not use this file except in compliance with the License.                        //
// You may obtain a copy of the License at                                                 //
//                                                                                         //
//   http://www.apache.org/licenses/LICENSE-2.0                                            //
//                                                                                         //
// Unless required by applicable law or agreed to in writing, software                     //
// distributed under the License is distributed on an "AS IS" BASIS,                       //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                //
// See the License for the specific language governing permissions and                     //
// limitations under the License.                                                          //
                                                                                           //
Meteor.publish("contactProfiles", function (showAll) {                                     // 17
  var db = this.connection.sandstormDb;                                                    // 18
  var _this = this;                                                                        // 19
                                                                                           //
  // We maintain a map from identity IDs to live query handles that track profile changes.
  var contactIdentities = {};                                                              // 22
  var disallowGuests = db.getOrganizationDisallowGuests();                                 // 23
                                                                                           //
  function addIdentityOfContact(contact) {                                                 // 25
    if (!(contact.identityId in contactIdentities)) {                                      // 26
      var user = Meteor.users.findOne({ _id: contact.identityId });                        // 27
                                                                                           //
      if (disallowGuests && !showAll) {                                                    // 29
        if (!db.isIdentityInOrganization(user)) {                                          // 30
          return;                                                                          // 31
        }                                                                                  //
      }                                                                                    //
                                                                                           //
      if (user) {                                                                          // 35
        SandstormDb.fillInProfileDefaults(user);                                           // 36
        SandstormDb.fillInIntrinsicName(user);                                             // 37
        var filteredUser = _.pick(user, "_id", "profile");                                 // 38
        _this.added("contactProfiles", user._id, filteredUser);                            // 39
      }                                                                                    //
                                                                                           //
      contactIdentities[contact.identityId] = Meteor.users.find({ _id: contact.identityId }, { fields: { profile: 1 } }).observeChanges({
        changed: function (id, fields) {                                                   // 44
          _this.changed("contactProfiles", id, fields);                                    // 45
        }                                                                                  //
      });                                                                                  //
    }                                                                                      //
  }                                                                                        //
                                                                                           //
  var cursor = db.collections.contacts.find({ ownerId: this.userId });                     // 51
                                                                                           //
  var handle = cursor.observe({                                                            // 53
    added: function (contact) {                                                            // 54
      addIdentityOfContact(contact);                                                       // 55
    },                                                                                     //
                                                                                           //
    changed: function (contact) {                                                          // 58
      addIdentityOfContact(contact);                                                       // 59
    },                                                                                     //
                                                                                           //
    removed: function (contact) {                                                          // 62
      _this.removed("contactProfiles", contact.identityId);                                // 63
      contactIdentities[contact.identityId].stop();                                        // 64
      delete contactIdentities[contact.identityId];                                        // 65
    }                                                                                      //
  });                                                                                      //
  this.ready();                                                                            // 68
                                                                                           //
  this.onStop(function () {                                                                // 70
    handle.stop();                                                                         // 71
    Object.keys(contactIdentities).forEach(function (identityId) {                         // 72
      contactIdentities[identityId].stop();                                                // 73
      delete contactIdentities[identityId];                                                // 74
    });                                                                                    //
  });                                                                                      //
});                                                                                        //
/////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-contacts'] = {
  ContactProfiles: ContactProfiles
};

})();

//# sourceMappingURL=sandstorm-contacts.js.map
